"""Compute operator components.

.. todo::

    This is the only part of the new runner making use of the global context
    through the :class:`EKO` object.

    After #242, the goal is to update :class:`Operator` and
    :class:`OperatorMatrixElement` to simplify the computation and passing down
    parameters.

"""
import numpy as np

from .. import evolution_operator as evop
from ..evolution_operator import matching_condition
from ..evolution_operator import operator_matrix_element as ome
from ..evolution_operator import physical
from ..io import EKO
from ..io.items import Evolution, Matching, Operator
from ..quantities.heavy_quarks import QuarkMassScheme
from . import commons


def managers(eko: EKO) -> dict:
    """Collect managers for operator computation.

    .. todo::

        Legacy interface, avoid managers usage.

    """
    tcard = eko.theory_card
    ocard = eko.operator_card
    return dict(
        thresholds_config=commons.atlas(tcard, ocard),
        couplings=commons.couplings(tcard, ocard),
        interpol_dispatcher=commons.interpolator(ocard),
    )


def blowup_info(eko: EKO) -> dict:
    """Prepare common information to blow up to flavor basis.

    Note
    ----
    ``intrinsic_range`` is a fully deprecated feature, here and anywhere else,
    since a full range is already always used for backward evolution, and it is
    not harmful to use it also for forward.

    Indeed, the only feature of non-intrinsic evolution is to absorb a
    non-trivial boundary condition when an intrinsic PDF is defined.
    But to achieve this, is sufficient to not specify any intrinsic boundary
    condition at all, while if something is there, it is intuitive enough that
    it will be consistently evolved.

    Moreover, since two different behavior are applied for the forward and
    backward evolution, the intrinsic range is a "non-local" function, since it
    does not depend only on the evolution segment, but also on the previous
    evolution history (to determine if evolution is backward in flavor,
    irrespectively of happening for an increasing or decreasing interval in
    scale at fixed flavor).

    """
    return dict(intrinsic_range=[4, 5, 6], qed=eko.theory_card.order[1] > 0)


def evolve_configs(eko: EKO) -> dict:
    """Create configs for :class:`Operator`.

    .. todo::

        Legacy interface, make use of a dedicated object.

    """
    tcard = eko.theory_card
    ocard = eko.operator_card
    return dict(
        order=tcard.order,
        xif2=tcard.xif**2,
        method=ocard.configs.evolution_method.value,
        ev_op_iterations=ocard.configs.ev_op_iterations,
        ev_op_max_order=ocard.configs.ev_op_max_order,
        polarized=ocard.configs.polarized,
        time_like=ocard.configs.time_like,
        debug_skip_singlet=ocard.debug.skip_singlet,
        debug_skip_non_singlet=ocard.debug.skip_non_singlet,
        n_integration_cores=ocard.configs.n_integration_cores,
        ModSV=ocard.configs.scvar_method,
        n3lo_ad_variation=tcard.n3lo_ad_variation,
    )


def evolve(eko: EKO, recipe: Evolution) -> Operator:
    """Compute evolution in isolation."""
    op = evop.Operator(
        evolve_configs(eko), managers(eko), recipe.as_atlas, is_threshold=recipe.cliff
    )
    op.compute()

    binfo = blowup_info(eko)
    res, err = physical.PhysicalOperator.ad_to_evol_map(
        op.op_members, op.nf, op.q2_to, **binfo
    ).to_flavor_basis_tensor(qed=binfo["qed"])

    return Operator(res, err)


def matching_configs(eko: EKO) -> dict:
    """Create configs for :class:`OperatorMatrixElement`.

    .. todo::

        Legacy interface, make use of a dedicated object.

    """
    tcard = eko.theory_card
    ocard = eko.operator_card
    return dict(
        **evolve_configs(eko),
        backward_inversion=ocard.configs.inversion_method,
        intrinsic_range=tcard.heavy.intrinsic_flavors,
    )


def match(eko: EKO, recipe: Matching) -> Operator:
    """Compute matching in isolation.

    Note
    ----
    Compared to the old prescription, a dedicated rotation to a different
    intrinsic basis is not needed any longer.

    All the operators are blown up to flavor basis, and they are saved and
    joined in that unique basis. So, the only rotation used is towards that
    basis, and encoded in the blowing up prescription.

    """
    kthr = eko.theory_card.heavy.squared_ratios[recipe.hq - 4]
    op = ome.OperatorMatrixElement(
        matching_configs(eko),
        managers(eko),
        recipe.hq - 1,
        recipe.scale,
        recipe.inverse,
        np.log(kthr),
        eko.theory_card.heavy.masses_scheme is QuarkMassScheme.MSBAR,
    )
    op.compute()

    binfo = blowup_info(eko)
    nf_match = op.nf - 1 if recipe.inverse else op.nf
    res, err = matching_condition.MatchingCondition.split_ad_to_evol_map(
        op.op_members, nf_match, recipe.scale, **binfo
    ).to_flavor_basis_tensor(qed=binfo["qed"])

    return Operator(res, err)



"""Mocking tools."""
import numpy as np


def eko_identity(shape):
    """Generate an identity operator with the given shape.

    The operator has to be squared in PIDs and x grid, thus the last two
    elements of `shape` are not read, and they are considered to be equal to the
    second and third element respectively.

    Parameters
    ----------
    shape: 5-tuple
        specifies the required output tuple (last two elements are ignored)

    Returns
    -------
    np.array
        rank 5 array, representing an identity operator of the requested shape

    """
    i, k = np.ogrid[: shape[1], : shape[2]]
    eko_identity = np.zeros(shape[1:], int)
    eko_identity[i, k, i, k] = 1
    return np.broadcast_to(eko_identity[np.newaxis, :, :, :, :], shape)



r"""The Altarelli-Parisi splitting kernels.

Normalization is given by

.. math::
    \mathbf{P}(x) = \sum\limits_{j=0} a_s^{j+1} \mathbf P^{(j)}(x)

with :math:`a_s = \frac{\alpha_S(\mu^2)}{4\pi}`.
The 3-loop references for the non-singlet :cite:`Moch:2004pa`
and singlet :cite:`Vogt:2004mw` case contain also the lower
order results. The results are also determined in Mellin space in
terms of the anomalous dimensions (note the additional sign!)

.. math::
    \gamma(N) = - \mathcal{M}[\mathbf{P}(x)](N)
"""

import numba as nb
import numpy as np


@nb.njit(cache=True)
def exp_matrix_2D(gamma_S):
    r"""Compute the exponential and the eigensystem of the singlet anomalous dimension matrix.

    Parameters
    ----------
    gamma_S : numpy.ndarray
        singlet anomalous dimension matrix

    Returns
    -------
    exp : numpy.ndarray
        exponential of the singlet anomalous dimension matrix :math:`\gamma_{S}(N)`
    lambda_p : complex
        positive eigenvalue of the singlet anomalous dimension matrix
        :math:`\gamma_{S}(N)`
    lambda_m : complex
        negative eigenvalue of the singlet anomalous dimension matrix
        :math:`\gamma_{S}(N)`
    e_p : numpy.ndarray
        projector for the positive eigenvalue of the singlet anomalous
        dimension matrix :math:`\gamma_{S}(N)`
    e_m : numpy.ndarray
        projector for the negative eigenvalue of the singlet anomalous
        dimension matrix :math:`\gamma_{S}(N)`

    """
    # compute eigenvalues
    det = np.sqrt(
        np.power(gamma_S[0, 0] - gamma_S[1, 1], 2) + 4.0 * gamma_S[0, 1] * gamma_S[1, 0]
    )
    lambda_p = 1.0 / 2.0 * (gamma_S[0, 0] + gamma_S[1, 1] + det)
    lambda_m = 1.0 / 2.0 * (gamma_S[0, 0] + gamma_S[1, 1] - det)
    # compute projectors
    identity = np.identity(2, np.complex_)
    c = 1.0 / det
    e_p = +c * (gamma_S - lambda_m * identity)
    e_m = -c * (gamma_S - lambda_p * identity)
    exp = e_m * np.exp(lambda_m) + e_p * np.exp(lambda_p)
    return exp, lambda_p, lambda_m, e_p, e_m


@nb.njit(cache=True)
def exp_matrix(gamma):
    r"""Compute the exponential and the eigensystem of a matrix.

    Parameters
    ----------
    gamma : numpy.ndarray
        input matrix

    Returns
    -------
    exp : numpy.ndarray
        exponential of the matrix gamma :math:`\gamma(N)`
    w : numpy.ndarray
        array of the eigenvalues of the matrix lambda
    e : numpy.ndarray
        projectors on the eigenspaces of the matrix gamma :math:`\gamma(N)`

    """
    dim = gamma.shape[0]
    e = np.zeros((dim, dim, dim), np.complex_)
    # if dim == 2:
    #     exp, lambda_p, lambda_m, e_p, e_m = exp_matrix_2D(gamma)
    #     e[0] = e_p
    #     e[1] = e_m
    #     return exp, np.array([lambda_p, lambda_m]), e
    w, v = np.linalg.eig(gamma)
    v_inv = np.linalg.inv(v)
    exp = np.zeros((dim, dim), np.complex_)
    for i in range(dim):
        e[i] = np.outer(v[:, i], v_inv[i])
        exp += e[i] * np.exp(w[i])
    return exp, w, e



"""Structures to hold runcards information.

All energy scales in the runcards should be saved linearly, not the squared
value, for consistency.
Squares are consistenly taken inside.

"""
from dataclasses import dataclass
from math import nan
from typing import List, Optional, Union

import numpy as np
import numpy.typing as npt

from .. import basis_rotation as br
from .. import interpolation, msbar_masses
from .. import version as vmod
from ..couplings import couplings_mod_ev
from ..matchings import Atlas, nf_default
from ..quantities import heavy_quarks as hq
from ..quantities.couplings import CouplingsInfo
from ..quantities.heavy_quarks import HeavyInfo, QuarkMassScheme
from .dictlike import DictLike
from .types import EvolutionMethod
from .types import EvolutionPoint as EPoint
from .types import (
    InversionMethod,
    N3LOAdVariation,
    Order,
    RawCard,
    ScaleVariationsMethod,
    SquaredScale,
    T,
)


# TODO: add frozen
@dataclass
class TheoryCard(DictLike):
    """Represent theory card content."""

    order: Order
    """Perturbative order tuple, ``(QCD, QED)``."""
    couplings: CouplingsInfo
    """Couplings configuration."""
    heavy: HeavyInfo
    """Heavy quarks related information."""
    xif: float
    """Ratio between factorization scale and process scale."""
    n3lo_ad_variation: N3LOAdVariation
    """|N3LO| anomalous dimension variation: ``(gg_var, gq_var, qg_var, qq_var)``."""


@dataclass
class Debug(DictLike):
    """Debug configurations."""

    skip_singlet: bool = False
    """Whether to skip QCD singlet computation."""
    skip_non_singlet: bool = False
    """Whether to skip QCD non-singlet computation."""


@dataclass
class Configs(DictLike):
    """Solution specific configurations."""

    evolution_method: EvolutionMethod
    """Evolution mode."""
    ev_op_max_order: Order
    """Maximum order to use in U matrices expansion.
    Used only in ``perturbative`` solutions.
    """
    ev_op_iterations: int
    """Number of intervals in which to break the global path."""
    scvar_method: Optional[ScaleVariationsMethod]
    """Scale variation method."""
    inversion_method: Optional[InversionMethod]
    """Which method to use for backward matching conditions."""
    interpolation_polynomial_degree: int
    """Degree of elements of the intepolation polynomial basis."""
    interpolation_is_log: bool
    r"""Whether to use polynomials in :math:`\log(x)`.
    If `false`, polynomials are in :math:`x`.
    """
    polarized: bool
    """If `true` do polarized evolution."""
    time_like: bool
    """If `true` do time-like evolution."""
    n_integration_cores: int = 1
    """Number of cores used to parallelize integration."""


@dataclass
class OperatorCard(DictLike):
    """Operator Card info."""

    mu0: float
    """Initial scale."""
    mugrid: List[EPoint]
    xgrid: interpolation.XGrid
    """Momentum fraction internal grid."""

    # collections
    configs: Configs
    """Specific configuration to be used during the calculation of these operators."""
    debug: Debug
    """Debug configurations."""

    # optional
    eko_version: str = vmod.__version__
    """Version of EKO package first used for generation."""

    # a few properties, for ease of use and compatibility
    @property
    def mu20(self):
        """Squared value of initial scale."""
        return self.mu0**2

    @property
    def mu2grid(self) -> npt.NDArray:
        """Grid of squared final scales."""
        return np.array([mu for mu, _ in self.mugrid]) ** 2

    @property
    def evolgrid(self) -> List[EPoint]:
        """Grid of squared final scales."""
        return [(mu**2, nf) for mu, nf in self.mugrid]

    @property
    def pids(self):
        """Internal flavor basis, used for computation."""
        return np.array(br.flavor_basis_pids)


Card = Union[TheoryCard, OperatorCard]


@dataclass
class Legacy:
    """Upgrade legacy runcards."""

    theory: RawCard
    operator: RawCard

    MOD_EV2METHOD = {
        "EXA": "iterate-exact",
        "EXP": "iterate-expanded",
        "TRN": "truncated",
    }

    @staticmethod
    def heavies(pattern: str, old_th: dict):
        """Retrieve a set of values for all heavy flavors."""
        return [old_th[pattern % q] for q in hq.FLAVORS]

    @staticmethod
    def fallback(*args: T, default: Optional[T] = None) -> T:
        """Return the first not None argument."""
        for arg in args:
            if arg is not None:
                return arg

        if default is None:
            return args[-1]
        return default

    @property
    def new_theory(self):
        """Build new format theory runcard."""
        old = self.theory
        new = {}

        new["order"] = [old["PTO"] + 1, old["QED"]]
        alphaem = self.fallback(old.get("alphaqed"), old.get("alphaem"), default=0.0)
        ms = self.heavies("m%s", old)
        ks = self.heavies("k%sThr", old)
        new["couplings"] = dict(
            alphas=old["alphas"],
            alphaem=alphaem,
            scale=old["Qref"],
            num_flavs_ref=old["nfref"],
            max_num_flavs=old["MaxNfAs"],
        )
        new["heavy"] = {
            "num_flavs_init": nf_default(old["Qref"] ** 2.0, default_atlas(ms, ks))
            if old["nf0"] is None
            else old["nf0"],
            "num_flavs_max_pdf": old["MaxNfPdf"],
            "matching_ratios": self.heavies("k%sThr", old),
            "masses_scheme": old["HQ"],
        }
        intrinsic = []
        for idx, q in enumerate(hq.FLAVORS):
            if old.get(f"i{q}".upper()) == 1:
                intrinsic.append(idx + 4)
        new["heavy"]["intrinsic_flavors"] = intrinsic
        if old["HQ"] == "POLE":
            new["heavy"]["masses"] = [[m, nan] for m in ms]
        elif old["HQ"] == "MSBAR":
            mus = self.heavies("Qm%s", old)
            new["heavy"]["masses"] = [[m, mu] for m, mu in zip(ms, mus)]
        else:
            raise ValueError(f"Unknown mass scheme '{old['HQ']}'")

        new["xif"] = old["XIF"]
        new["n3lo_ad_variation"] = old.get("n3lo_ad_variation", (0, 0, 0, 0))

        return TheoryCard.from_dict(new)

    @property
    def new_operator(self):
        """Build new format operator runcard."""
        old = self.operator
        old_th = self.theory
        new = {}

        new["mu0"] = old_th["Q0"]
        if "mugrid" in old:
            mugrid = old["mugrid"]
        else:
            mu2grid = old["Q2grid"] if "Q2grid" in old else old["mu2grid"]
            mugrid = np.sqrt(mu2grid)
        new["mugrid"] = flavored_mugrid(
            mugrid,
            list(self.heavies("m%s", old_th)),
            list(self.heavies("k%sThr", old_th)),
        )

        new["configs"] = {}
        evmod = old_th["ModEv"]
        new["configs"]["evolution_method"] = self.MOD_EV2METHOD.get(evmod, evmod)
        new["configs"]["inversion_method"] = old_th.get(
            "backward_inversion", "expanded"
        )
        new["configs"]["scvar_method"] = old_th.get("ModSV", "expanded")
        for k in (
            "interpolation_polynomial_degree",
            "interpolation_is_log",
            "ev_op_iterations",
            "n_integration_cores",
            "polarized",
            "time_like",
        ):
            new["configs"][k] = old[k]
        max_order = old["ev_op_max_order"]
        if isinstance(max_order, int):
            new["configs"]["ev_op_max_order"] = [max_order, old_th["QED"]]

        new["debug"] = {}
        lpref = len("debug_")
        for k in ("debug_skip_non_singlet", "debug_skip_singlet"):
            new["debug"][k[lpref:]] = old[k]

        new["xgrid"] = old["interpolation_xgrid"]

        return OperatorCard.from_dict(new)


def update(theory: Union[RawCard, TheoryCard], operator: Union[RawCard, OperatorCard]):
    """Update legacy runcards.

    This function is mainly defined for compatibility with the old interface.
    Prefer direct usage of :class:`Legacy` in new code.

    Consecutive applications of this function yield identical results::

        cards = update(theory, operator)
        assert update(*cards) == cards

    """
    if isinstance(theory, TheoryCard) or isinstance(operator, OperatorCard):
        # if one is not a dict, both have to be new cards
        assert isinstance(theory, TheoryCard)
        assert isinstance(operator, OperatorCard)
        return theory, operator

    cards = Legacy(theory, operator)
    return cards.new_theory, cards.new_operator


def default_atlas(masses: list, matching_ratios: list):
    r"""Create default landscape.

    This method should not be used to write new runcards, but rather to have a
    consistent default for comparison with other softwares and existing PDF
    sets.
    There is no one-to-one relation between number of running flavors and final
    scales, unless matchings are all applied. But this is a custom choice,
    since it is possible to have PDFs in different |FNS| at the same scales.

    """
    matchings = (np.array(masses) * np.array(matching_ratios)) ** 2
    return Atlas(matchings.tolist(), (0.0, 0))


def flavored_mugrid(mugrid: list, masses: list, matching_ratios: list):
    r"""Upgrade :math:`\mu^2` grid to contain also target number flavors.

    It determines the number of flavors for the PDF set at the target scale,
    inferring it according to the specified scales.

    This method should not be used to write new runcards, but rather to have a
    consistent default for comparison with other softwares and existing PDF
    sets.
    There is no one-to-one relation between number of running flavors and final
    scales, unless matchings are all applied. But this is a custom choice,
    since it is possible to have PDFs in different |FNS| at the same scales.

    """
    atlas = default_atlas(masses, matching_ratios)
    return [(mu, nf_default(mu**2, atlas)) for mu in mugrid]


# TODO: move to a more suitable place
def masses(theory: TheoryCard, evmeth: EvolutionMethod) -> List[SquaredScale]:
    """Compute masses in the chosen scheme."""
    if theory.heavy.masses_scheme is QuarkMassScheme.MSBAR:
        return msbar_masses.compute(
            theory.heavy.masses,
            theory.couplings,
            theory.order,
            couplings_mod_ev(evmeth),
            np.power(theory.heavy.matching_ratios, 2.0).tolist(),
            xif2=theory.xif**2,
        ).tolist()
    if theory.heavy.masses_scheme is QuarkMassScheme.POLE:
        return [mq.value**2 for mq in theory.heavy.masses]

    raise ValueError(f"Unknown mass scheme '{theory.heavy.masses_scheme}'")



"""ekomark specialization of the navigator."""
import argparse
import pathlib

from banana import cfg as banana_cfg
from banana import navigator as bnav

from . import glob, navigator


def yelp(*args):
    """Help function (renamed to avoid clash of names) - short cut: h."""
    if len(args) == 0:
        print(
            f"""Welcome to ekomark navigator - the eko benchmark skript!
Available variables:
    {bnav.help_vars}
    o = "{bnav.o}" -> query operator
Available functions:
    {bnav.help_fncs}
    dfl(id) - log as DataFrame
    simlogs(id) - find similar logs
    diff(id,id) - subtractig logs
    check_log(id) - check logs passed
    crashed_log(id) - print crashed logs
    plot_pdfs(id) - plot pdfs to final scale
    display_pdfs(id) - open pdfs plots
    compare(id,id) - compare externals
"""
        )
    elif len(args) == 1:
        return help(*args)
    return None


def register_globals(configpath):
    """Expose global variables."""
    app = navigator.NavigatorApp(configpath, "sandbox")
    glob.app = app

    glob.glob["yelp"] = yelp
    glob.glob["h"] = yelp

    # register banana functions
    bnav.register_globals(glob.glob, glob.app)

    # add my functions
    glob.glob["check_log"] = app.check_log
    glob.glob["plot_pdfs"] = app.plot_pdfs
    glob.glob["display_pdfs"] = app.display_pdfs


def launch_navigator():
    """CLI Entry point."""
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-c", "--config", type=pathlib.Path, default=None, help="Path to config file"
    )

    args = parser.parse_args()

    register_globals(banana_cfg.detect(args.config))

    # ekomark.navigator makes the globals here (e.g. app, ls, t) available inside IPython
    return bnav.launch_navigator(
        ["eko", "ekomark", "ekomark.navigator.glob"], skip_cfg=True
    )



"""The |OME| for the non-trivial matching conditions in the |VFNS| evolution."""

import copy
import functools
import logging

import numba as nb
import numpy as np

import ekore.operator_matrix_elements.polarized.space_like as ome_ps
import ekore.operator_matrix_elements.unpolarized.space_like as ome_us
import ekore.operator_matrix_elements.unpolarized.time_like as ome_ut

from .. import basis_rotation as br
from .. import scale_variations as sv
from ..io.types import InversionMethod
from ..matchings import Segment
from . import Operator, QuadKerBase

logger = logging.getLogger(__name__)


@nb.njit(cache=True)
def build_ome(A, matching_order, a_s, backward_method):
    r"""Construct the matching expansion in :math:`a_s` with the appropriate method.

    Parameters
    ----------
    A : numpy.ndarray
        list of |OME|
    matching_order : tuple(int,int)
        perturbation matching order
    a_s : float
        strong coupling, needed only for the exact inverse
    backward_method : InversionMethod or None
        empty or method for inverting the matching condition (exact or expanded)

    Returns
    -------
    ome : numpy.ndarray
        matching operator matrix

    """
    # to get the inverse one can use this FORM snippet
    # Symbol a;
    # NTensor c,d,e;
    # Local x=-(a*c+a**2* d + a**3 * e);
    # Local bi = 1+x+x**2+x**3;
    # Print;
    # .end
    ome = np.eye(len(A[0]), dtype=np.complex_)
    A = A[:, :, :]
    A = np.ascontiguousarray(A)
    if backward_method is InversionMethod.EXPANDED:
        # expended inverse
        if matching_order[0] >= 1:
            ome -= a_s * A[0]
        if matching_order[0] >= 2:
            ome += a_s**2 * (-A[1] + A[0] @ A[0])
        if matching_order[0] >= 3:
            ome += a_s**3 * (-A[2] + A[0] @ A[1] + A[1] @ A[0] - A[0] @ A[0] @ A[0])
    else:
        # forward or exact inverse
        if matching_order[0] >= 1:
            ome += a_s * A[0]
        if matching_order[0] >= 2:
            ome += a_s**2 * A[1]
        if matching_order[0] >= 3:
            ome += a_s**3 * A[2]
        # need inverse exact ?  so add the missing pieces
        if backward_method is InversionMethod.EXACT:
            ome = np.linalg.inv(ome)
    return ome


@nb.njit(cache=True)
def quad_ker(
    u,
    order,
    mode0,
    mode1,
    is_log,
    logx,
    areas,
    a_s,
    nf,
    L,
    sv_mode,
    Lsv,
    backward_method,
    is_msbar,
    is_polarized,
    is_time_like,
):
    r"""Raw kernel inside quad.

    Parameters
    ----------
    u : float
        quad argument
    order : tuple(int,int)
        perturbation matching order
    mode0 : int
        pid for first element in the singlet sector
    mode1 : int
        pid for second element in the singlet sector
    is_log : boolean
        logarithmic interpolation
    logx : float
        Mellin inversion point
    areas : tuple
        basis function configuration
    a_s : float
        strong coupling, needed only for the exact inverse
    nf: int
        number of active flavor below threshold
    L : float
        :math:``\ln(\mu_F^2 / m_h^2)``
    backward_method : InversionMethod or None
        empty or method for inverting the matching condition (exact or expanded)
    is_msbar: bool
        add the |MSbar| contribution
    is_polarized : boolean
        is polarized evolution ?
    is_time_like : boolean
        is time-like evolution ?

    Returns
    -------
    ker : float
        evaluated integration kernel

    """
    ker_base = QuadKerBase(u, is_log, logx, mode0)
    integrand = ker_base.integrand(areas)
    if integrand == 0.0:
        return 0.0
    # compute the ome
    if ker_base.is_singlet or ker_base.is_QEDsinglet:
        indices = {21: 0, 100: 1, 90: 2}
        if is_polarized:
            if is_time_like:
                raise NotImplementedError("Polarized, time-like is not implemented")
            A = ome_ps.A_singlet(order, ker_base.n, nf, L)
        else:
            if is_time_like:
                A = ome_ut.A_singlet(order, ker_base.n, L)
            else:
                A = ome_us.A_singlet(order, ker_base.n, nf, L, is_msbar)
    else:
        indices = {200: 0, 91: 1}
        if is_polarized:
            if is_time_like:
                raise NotImplementedError("Polarized, time-like is not implemented")
            A = ome_ps.A_non_singlet(order, ker_base.n, L)
        else:
            if is_time_like:
                A = ome_ut.A_non_singlet(order, ker_base.n, L)
            else:
                A = ome_us.A_non_singlet(order, ker_base.n, nf, L)

    # correct for scale variations
    if sv_mode == sv.Modes.exponentiated:
        A = sv.exponentiated.gamma_variation(A, order, nf, Lsv)

    # build the expansion in alpha_s depending on the strategy
    ker = build_ome(A, order, a_s, backward_method)

    # select the needed matrix element
    ker = ker[indices[mode0], indices[mode1]]

    # recombine everything
    return np.real(ker * integrand)


class OperatorMatrixElement(Operator):
    r"""
    Internal representation of a single |OME|.

    The actual matrices are computed upon calling :meth:`compute`.

    Parameters
    ----------
    config : dict
        configuration
    managers : dict
        managers
    segment: Segment
        path segment
    L: float
        :math:`\ln(\mu_F^2 / m_h^2)`
    is_msbar: bool
        add the |MSbar| contribution
    """

    log_label = "Matching"
    # complete list of possible matching operators labels
    full_labels = [
        *br.singlet_labels,
        (br.matching_hplus_pid, 21),
        (br.matching_hplus_pid, 100),
        (21, br.matching_hplus_pid),
        (100, br.matching_hplus_pid),
        (br.matching_hplus_pid, br.matching_hplus_pid),
        (200, 200),
        (200, br.matching_hminus_pid),
        (br.matching_hminus_pid, 200),
        (br.matching_hminus_pid, br.matching_hminus_pid),
    ]
    # still valid in QED since Sdelta and Vdelta matchings are diagonal
    full_labels_qed = copy.deepcopy(full_labels)

    def __init__(self, config, managers, nf, q2, is_backward, L, is_msbar):
        super().__init__(config, managers, Segment(q2, q2, nf))
        self.backward_method = config["backward_inversion"] if is_backward else None
        if is_backward:
            self.is_intrinsic = True
        else:
            self.is_intrinsic = bool(len(config["intrinsic_range"]) != 0)
        self.L = L
        self.is_msbar = is_msbar
        # Note for the moment only QCD matching is implemented
        self.order = (self.order[0] - 1, self.order[1])

    @property
    def labels(self):
        """Necessary sector labels to compute.

        Returns
        -------
        list(str)
            sector labels
        """
        labels = []
        # non-singlet labels
        if self.config["debug_skip_non_singlet"]:
            logger.warning("%s: skipping non-singlet sector", self.log_label)
        else:
            labels.append((200, 200))
            if self.is_intrinsic or self.backward_method is not None:
                # intrinsic labels, which are not zero at NLO
                labels.append((br.matching_hminus_pid, br.matching_hminus_pid))
                # These contributions are always 0 for the moment
                # labels.extend([(200, br.matching_hminus_pid), (br.matching_hminus_pid, 200)])
        # same for singlet
        if self.config["debug_skip_singlet"]:
            logger.warning("%s: skipping singlet sector", self.log_label)
        else:
            labels.extend(
                [
                    *br.singlet_labels,
                    (br.matching_hplus_pid, 21),
                    (br.matching_hplus_pid, 100),
                ]
            )
            if self.is_intrinsic or self.backward_method is not None:
                labels.extend(
                    [
                        (21, br.matching_hplus_pid),
                        (100, br.matching_hplus_pid),
                        (br.matching_hplus_pid, br.matching_hplus_pid),
                    ]
                )
        return labels

    def quad_ker(self, label, logx, areas):
        """Return partially initialized integrand function.

        Parameters
        ----------
        label: tuple
            operator element pids
        logx: float
            Mellin inversion point
        areas : tuple
            basis function configuration

        Returns
        -------
        functools.partial
            partially initialized integration kernel
        """
        return functools.partial(
            quad_ker,
            order=self.order,
            mode0=label[0],
            mode1=label[1],
            is_log=self.int_disp.log,
            logx=logx,
            areas=areas,
            a_s=self.a_s,
            nf=self.nf,
            L=self.L,
            sv_mode=self.sv_mode,
            Lsv=np.log(self.xif2),
            backward_method=self.backward_method,
            is_msbar=self.is_msbar,
            is_polarized=self.config["polarized"],
            is_time_like=self.config["time_like"],
        )

    @property
    def a_s(self):
        """Return the computed values for :math:`a_s`.

        Note that here you need to use :math:`a_s^{n_f+1}`
        """
        sc = self.managers["couplings"]
        return sc.a_s(
            self.q2_from
            * (self.xif2 if self.sv_mode == sv.Modes.exponentiated else 1.0),
            nf_to=self.nf + 1,
        )

    def compute(self):
        """Compute the actual operators (i.e. run the integrations)."""
        self.initialize_op_members()

        # At LO you don't need anything else
        if self.order[0] == 0:
            logger.info("%s: no need to compute matching at LO", self.log_label)
            return

        self.integrate()



r"""The unpolarized, time-like |OME|."""

import numba as nb
import numpy as np

from . import as1


@nb.njit(cache=True)
def A_non_singlet(matching_order, _N, _L):
    r"""Compute the non-singlet |OME|.

    Parameters
    ----------
    matching_order : tuple(int, int)
        perturbative matching order

    Returns
    -------
    numpy.ndarray
        non-singlet |OME|

    """
    A_ns = np.zeros((matching_order[0], 2, 2), np.complex_)
    A_ns[0] = as1.A_ns()
    return A_ns


@nb.njit(cache=True)
def A_singlet(matching_order, N, L):
    r"""Compute the singlet |OME|.

    Parameters
    ----------
    matching_order : tuple(int, int)
        perturbative matching order
    N : complex
        Mellin moment
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    numpy.ndarray
        singlet |OME|

    """
    A_s = np.zeros((matching_order[0], 3, 3), np.complex_)
    A_s[0] = as1.A_singlet(N, L)
    return A_s



"""Compute the |NLO| Altarelli-Parisi splitting kernels.

These expression have been obtained using the procedure described in the
`wiki <https://github.com/N3PDF/eko/wiki/Parse-NLO-expressions>`_
involving ``FormGet`` :cite:`Hahn:2016ebn`.
"""

import numba as nb
import numpy as np

from eko import constants
from eko.constants import log2, zeta2, zeta3

from ....harmonics import cache as c


@nb.njit(cache=True)
def gamma_nsm(n, nf, cache):
    r"""Compute the |NLO| valence-like non-singlet anomalous dimension.

    Implements Eq. (3.5) of :cite:`Moch:2004pa`.

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        |NLO| valence-like non-singlet anomalous dimension
        :math:`\\gamma_{ns,-}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    # Here, Sp refers to S' ("s-prime") (german: "s-strich" or in Pegasus language: SSTR)
    # of :cite:`Gluck:1989ze` and NOT to the Spence function a.k.a. dilogarithm
    # TODO : these harmonic sums are computed also for the QED sector then we can use
    # the ones that are passed to the O(as1aem1) anomalous dimensions
    Sp1m = c.get(c.S1mh, cache, n)
    Sp2m = c.get(c.S2mh, cache, n)
    Sp3m = c.get(c.S3mh, cache, n)
    g3n = c.get(c.g3, cache, n)
    # fmt: off
    gqq1m_cfca = 16*g3n - (144 + n*(1 + n)*(156 + n*(340 + n*(655 + 51*n*(2 + n)))))/(18.*np.power(n,3)*np.power(1 + n,3)) + (-14.666666666666666 + 8/n - 8/(1 + n))*S2 - (4*Sp2m)/(n + np.power(n,2)) + S1*(29.77777777777778 + 16/np.power(n,2) - 16*S2 + 8*Sp2m) + 2*Sp3m + 10*zeta3 + zeta2*(16*S1 - 16*Sp1m - (16*(1 + n*log2))/n) # pylint: disable=line-too-long
    gqq1m_cfcf = -32*g3n + (24 - n*(-32 + 3*n*(-8 + n*(3 + n)*(3 + np.power(n,2)))))/(2.*np.power(n,3)*np.power(1 + n,3)) + (12 - 8/n + 8/(1 + n))*S2 + S1*(-24/np.power(n,2) - 8/np.power(1 + n,2) + 16*S2 - 16*Sp2m) + (8*Sp2m)/(n + np.power(n,2)) - 4*Sp3m - 20*zeta3 + zeta2*(-32*S1 + 32*Sp1m + 32*(1/n + log2)) # pylint: disable=line-too-long
    gqq1m_cfnf = (-12 + n*(20 + n*(47 + 3*n*(2 + n))))/(9.*np.power(n,2)*np.power(1 + n,2)) - (40*S1)/9. + (8*S2)/3. # pylint: disable=line-too-long
    # fmt: on
    result = constants.CF * (
        (constants.CA * gqq1m_cfca)
        + (constants.CF * gqq1m_cfcf)
        + (2.0 * constants.TR * nf * gqq1m_cfnf)
    )
    return result


@nb.njit(cache=True)
def gamma_nsp(n, nf, cache):
    r"""Compute the |NLO| singlet-like non-singlet anomalous dimension.

    Implements Eq. (3.5) of :cite:`Moch:2004pa`.

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        |NLO| singlet-like non-singlet anomalous dimension
        :math:`\\gamma_{ns,+}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    Sp1p = c.get(c.S1h, cache, n)
    Sp2p = c.get(c.S2h, cache, n)
    Sp3p = c.get(c.S3h, cache, n)
    g3n = c.get(c.g3, cache, n)
    # fmt: off
    gqq1p_cfca = -16*g3n + (132 - n*(340 + n*(655 + 51*n*(2 + n))))/(18.*np.power(n,2)*np.power(1 + n,2)) + (-14.666666666666666 + 8/n - 8/(1 + n))*S2 - (4*Sp2p)/(n + np.power(n,2)) + S1*(29.77777777777778 - 16/np.power(n,2) - 16*S2 + 8*Sp2p) + 2*Sp3p + 10*zeta3 + zeta2*(16*S1 - 16*Sp1p + 16*(1/n - log2)) # pylint: disable=line-too-long
    gqq1p_cfcf = 32*g3n - (8 + n*(32 + n*(40 + 3*n*(3 + n)*(3 + np.power(n,2)))))/(2.*np.power(n,3)*np.power(1 + n,3)) + (12 - 8/n + 8/(1 + n))*S2 + S1*(40/np.power(n,2) - 8/np.power(1 + n,2) + 16*S2 - 16*Sp2p) + (8*Sp2p)/(n + np.power(n,2)) - 4*Sp3p - 20*zeta3 + zeta2*(-32*S1 + 32*Sp1p + 32*(-(1/n) + log2)) # pylint: disable=line-too-long
    gqq1p_cfnf = (-12 + n*(20 + n*(47 + 3*n*(2 + n))))/(9.*np.power(n,2)*np.power(1 + n,2)) - (40*S1)/9. + (8*S2)/3. # pylint: disable=line-too-long
    # fmt: on
    result = constants.CF * (
        (constants.CA * gqq1p_cfca)
        + (constants.CF * gqq1p_cfcf)
        + (2.0 * constants.TR * nf * gqq1p_cfnf)
    )
    return result


@nb.njit(cache=True)
def gamma_ps(n, nf):
    r"""Compute the |NLO| pure-singlet quark-quark anomalous dimension.

    Implements Eq. (3.6) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors

    Returns
    -------
    complex
        |NLO| pure-singlet quark-quark anomalous dimension
        :math:`\\gamma_{ps}^{(1)}(N)`

    """
    # fmt: off
    gqqps1_nfcf = (-4*(2 + n*(5 + n))*(4 + n*(4 + n*(7 + 5*n))))/((-1 + n)*np.power(n,3)*np.power(1 + n,3)*np.power(2 + n,2)) # pylint: disable=line-too-long
    # fmt: on
    result = 2.0 * constants.TR * nf * constants.CF * gqqps1_nfcf
    return result


@nb.njit(cache=True)
def gamma_qg(n, nf, cache):
    r"""Compute the |NLO| quark-gluon singlet anomalous dimension.

    Implements Eq. (3.7) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        |NLO| quark-gluon singlet anomalous dimension
        :math:`\\gamma_{qg}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    Sp2p = c.get(c.S2h, cache, n)
    # fmt: off
    gqg1_nfca = (-4*(16 + n*(64 + n*(104 + n*(128 + n*(85 + n*(36 + n*(25 + n*(15 + n*(6 + n))))))))))/((-1 + n)*np.power(n,3)*np.power(1 + n,3)*np.power(2 + n,3)) - (16*(3 + 2*n)*S1)/np.power(2 + 3*n + np.power(n,2),2) + (4*(2 + n + np.power(n,2))*np.power(S1,2))/(n*(2 + 3*n + np.power(n,2))) - (4*(2 + n + np.power(n,2))*S2)/(n*(2 + 3*n + np.power(n,2))) + (4*(2 + n + np.power(n,2))*Sp2p)/(n*(2 + 3*n + np.power(n,2))) # pylint: disable=line-too-long
    gqg1_nfcf = (-2*(4 + n*(8 + n*(1 + n)*(25 + n*(26 + 5*n*(2 + n))))))/(np.power(n,3)*np.power(1 + n,3)*(2 + n)) + (8*S1)/np.power(n,2) - (4*(2 + n + np.power(n,2))*np.power(S1,2))/(n*(2 + 3*n + np.power(n,2))) + (4*(2 + n + np.power(n,2))*S2)/(n*(2 + 3*n + np.power(n,2))) # pylint: disable=line-too-long
    # fmt: on
    result = (
        2.0 * constants.TR * nf * (constants.CA * gqg1_nfca + constants.CF * gqg1_nfcf)
    )
    return result


@nb.njit(cache=True)
def gamma_gq(n, nf, cache):
    r"""Compute the |NLO| gluon-quark singlet anomalous dimension.

    Implements Eq. (3.8) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        |NLO| gluon-quark singlet anomalous dimension
        :math:`\\gamma_{gq}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    Sp2p = c.get(c.S2h, cache, n)
    # fmt: off
    ggq1_cfcf = (-8 + 2*n*(-12 + n*(-1 + n*(28 + n*(43 + 6*n*(5 + 2*n))))))/((-1 + n)*np.power(n,3)*np.power(1 + n,3)) - (4*(10 + n*(17 + n*(8 + 5*n)))*S1)/((-1 + n)*n*np.power(1 + n,2)) + (4*(2 + n + np.power(n,2))*np.power(S1,2))/(n*(-1 + np.power(n,2))) + (4*(2 + n + np.power(n,2))*S2)/(n*(-1 + np.power(n,2))) # pylint: disable=line-too-long
    ggq1_cfca = (-4*(144 + n*(432 + n*(-152 + n*(-1304 + n*(-1031 + n*(695 + n*(1678 + n*(1400 + n*(621 + 109*n))))))))))/(9.*np.power(n,3)*np.power(1 + n,3)*np.power(-2 + n + np.power(n,2),2)) + (4*(-12 + n*(-22 + 41*n + 17*np.power(n,3)))*S1)/(3.*np.power(-1 + n,2)*np.power(n,2)*(1 + n)) + ((8 + 4*n + 4*np.power(n,2))*np.power(S1,2))/(n - np.power(n,3)) + ((8 + 4*n + 4*np.power(n,2))*S2)/(n - np.power(n,3)) + (4*(2 + n + np.power(n,2))*Sp2p)/(n*(-1 + np.power(n,2))) # pylint: disable=line-too-long
    ggq1_cfnf = (8*(16 + n*(27 + n*(13 + 8*n))))/(9.*(-1 + n)*n*np.power(1 + n,2)) - (8*(2 + n + np.power(n,2))*S1)/(3.*n*(-1 + np.power(n,2))) # pylint: disable=line-too-long
    # fmt: on
    result = constants.CF * (
        (constants.CA * ggq1_cfca)
        + (constants.CF * ggq1_cfcf)
        + (2.0 * constants.TR * nf * ggq1_cfnf)
    )
    return result


@nb.njit(cache=True)
def gamma_gg(n, nf, cache):
    r"""Compute the |NLO| gluon-gluon singlet anomalous dimension.

    Implements Eq. (3.9) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        |NLO| gluon-gluon singlet anomalous dimension
        :math:`\\gamma_{gg}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, n)
    Sp1p = c.get(c.S1h, cache, n)
    Sp2p = c.get(c.S2h, cache, n)
    Sp3p = c.get(c.S3h, cache, n)
    g3n = c.get(c.g3, cache, n)
    # fmt: off
    ggg1_caca = 16*g3n - (2*(576 + n*(1488 + n*(560 + n*(-1248 + n*(-1384 + n*(1663 + n*(4514 + n*(4744 + n*(3030 + n*(1225 + 48*n*(7 + n))))))))))))/(9.*np.power(-1 + n,2)*np.power(n,3)*np.power(1 + n,3)*np.power(2 + n,3)) + S1*(29.77777777777778 + 16/np.power(-1 + n,2) + 16/np.power(1 + n,2) - 16/np.power(2 + n,2) - 8*Sp2p) + (16*(1 + n + np.power(n,2))*Sp2p)/(n*(1 + n)*(-2 + n + np.power(n,2))) - 2*Sp3p - 10*zeta3 + zeta2*(-16*S1 + 16*Sp1p + 16*(-(1/n) + log2)) # pylint: disable=line-too-long
    ggg1_canf = (8*(6 + n*(1 + n)*(28 + n*(1 + n)*(13 + 3*n*(1 + n)))))/(9.*np.power(n,2)*np.power(1 + n,2)*(-2 + n + np.power(n,2))) - (40*S1)/9. # pylint: disable=line-too-long
    ggg1_cfnf = (2*(-8 + n*(-8 + n*(-10 + n*(-22 + n*(-3 + n*(6 + n*(8 + n*(4 + n)))))))))/(np.power(n,3)*np.power(1 + n,3)*(-2 + n + np.power(n,2))) # pylint: disable=line-too-long
    # fmt: on
    result = constants.CA * constants.CA * ggg1_caca + 2.0 * constants.TR * nf * (
        constants.CA * ggg1_canf + constants.CF * ggg1_cfnf
    )
    return result


@nb.njit(cache=True)
def gamma_singlet(n, nf, cache):
    r"""Compute the next-leading-order singlet anomalous dimension matrix.

    .. math::
        \\gamma_S^{(1)} = \\left(\begin{array}{cc}
        \\gamma_{qq}^{(1)} & \\gamma_{qg}^{(1)}\\
        \\gamma_{gq}^{(1)} & \\gamma_{gg}^{(1)}
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        |NLO| singlet anomalous dimension matrix :math:`\\gamma_{S}^{(1)}(N)`

    """
    gamma_qq = gamma_nsp(n, nf, cache) + gamma_ps(n, nf)
    gamma_S_0 = np.array(
        [
            [gamma_qq, gamma_qg(n, nf, cache)],
            [gamma_gq(n, nf, cache), gamma_gg(n, nf, cache)],
        ],
        np.complex_,
    )
    return gamma_S_0


@nb.njit(cache=True)
def gamma_singlet_qed(N, nf, cache):
    r"""Compute the leading-order singlet anomalous dimension matrix for the unified evolution basis.

    .. math::
        \\gamma_S^{(2,0)} = \\left(\begin{array}{cccc}
        \\gamma_{gg}^{(2,0)} & 0 & \\gamma_{gq}^{(2,0)} & 0\\
        0 & 0 & 0 & 0 \\
        \\gamma_{qg}^{(2,0)} & 0 & \\gamma_{qq}^{(2,0)} & 0 \\
        0 & 0 & 0 & \\gamma_{qq}^{(2,0)} \\
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{S}^{(2,0)}(N)`

    """
    gamma_ns_p = gamma_nsp(N, nf, cache)
    gamma_qq = gamma_ns_p + gamma_ps(N, nf)
    gamma_S = np.array(
        [
            [gamma_gg(N, nf, cache), 0.0 + 0.0j, gamma_gq(N, nf, cache), 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            [gamma_qg(N, nf, cache), 0.0 + 0.0j, gamma_qq, 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, gamma_ns_p],
        ],
        np.complex_,
    )
    return gamma_S


@nb.njit(cache=True)
def gamma_valence_qed(N, nf, cache):
    r"""Compute the leading-order valence anomalous dimension matrix for the unified evolution basis.

    .. math::
        \\gamma_V^{(2,0)} = \\left(\begin{array}{cc}
        \\gamma_{ns-}^{(2,0)} & 0\\
        0 & \\gamma_{ns-}^{(2,0)}
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{V}^{(2,0)}(N)`

    """
    gamma_V = np.array(
        [
            [1.0, 0.0],
            [0.0, 1.0],
        ],
        np.complex_,
    )
    return gamma_V * gamma_nsm(N, nf, cache)



"""Launch EKO calculations, with legacy mu2grid mode."""
import pathlib
from typing import Sequence

import click
import yaml

import eko
from eko.io.runcards import OperatorCard, TheoryCard

from .base import command
from .library import OPERATOR, OUTPUT, THEORY


@command.command("run")
@click.argument("paths", nargs=-1, type=click.Path(path_type=pathlib.Path))
def subcommand(paths: Sequence[pathlib.Path]):
    r"""Launch calculation specified by PATHS.

    Compute the operator given runcards. How they are specified depends on the
    amount of arguments:

        1. only one argument provided: it is used as the folder where runcards
        and output are located (with default names)

        2. two arguments: they are used as theory and operator card paths, and
        the output is placed in the same folder of the operator card

        3. three arguments: same as two, but the third argument is used as
        output path

    \b
    Default names are:
    - theory card: "theory.yaml"
    - operator card: "operator.yaml"
    - output: "eko.tar".

    """
    if len(paths) == 1:
        theory = paths[0] / THEORY
        operator = paths[0] / OPERATOR
    elif len(paths) in [2, 3]:
        theory = paths[0]
        operator = paths[1]
    else:
        raise click.UsageError(
            f"Only one to three arguments allowed, '{len(paths)}' passed."
        )

    if len(paths) == 3:
        output = paths[2]
    else:
        output = operator.parent / OUTPUT

    theory = yaml.safe_load(theory.read_text(encoding="utf-8"))
    if "order" in theory:
        theory = TheoryCard.from_dict(theory)
    operator = yaml.safe_load(operator.read_text(encoding="utf-8"))
    if "configs" in operator:
        operator = OperatorCard.from_dict(operator)

    eko.solve(theory, operator, path=output)



"""The :math:`O(a_{em}^2)` Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np

from eko import constants

from ....harmonics import cache as c
from . import as1aem1


@nb.njit(cache=True)
def gamma_phph(N, nf):
    r"""Compute the :math:`O(a_{em}^2)` photon-photon singlet anomalous dimension.

    Implements Eq. (68) of :cite:`deFlorian:2016gvk`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` photon-photon singlet anomalous dimension
        :math:`\\gamma_{\\gamma \\gamma}^{(0,2)}(N)`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return (
        constants.NC
        * (nu * constants.eu2**2 + nd * constants.ed2**2)
        * (as1aem1.gamma_gph(N) / constants.CF / constants.CA + 4)
    )


@nb.njit(cache=True)
def gamma_uph(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` quark-photon anomalous dimension for up quarks.

    Implements Eq. (55) of :cite:`deFlorian:2016gvk` for q=u.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` quark-photon anomalous dimension :math:`\\gamma_{u \\gamma}^{(0,2)}(N)`

    """
    return constants.eu2 * as1aem1.gamma_qph(N, nf, cache) / constants.CF


@nb.njit(cache=True)
def gamma_dph(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` quark-photon anomalous dimension for down quarks.

    Implements Eq. (55) of :cite:`deFlorian:2016gvk` for q=d.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` quark-photon anomalous dimension :math:`\\gamma_{d \\gamma}^{(0,2)}(N)`

    """
    return constants.ed2 * as1aem1.gamma_qph(N, nf, cache) / constants.CF


@nb.njit(cache=True)
def gamma_phu(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` photon-quark anomalous dimension for up quarks.

    Implements Eq. (56) of :cite:`deFlorian:2016gvk` for q=u.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` photon-quark anomalous dimension :math:`\\gamma_{\\gamma u}^{(0,2)}(N)`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    S1 = c.get(c.S1, cache, N)
    tmp = (-16.0 * (-16.0 - 27.0 * N - 13.0 * N**2 - 8.0 * N**3)) / (
        9.0 * (-1.0 + N) * N * (1.0 + N) ** 2
    ) - 16.0 * (2.0 + 3.0 * N + 2.0 * N**2 + N**3) / (
        3.0 * (-1.0 + N) * N * (1.0 + N) ** 2
    ) * S1
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    return constants.eu2 * as1aem1.gamma_phq(N, cache) / constants.CF + eSigma2 * tmp


@nb.njit(cache=True)
def gamma_phd(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` photon-quark anomalous dimension for down quarks.

    Implements Eq. (56) of :cite:`deFlorian:2016gvk` for q=d.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` photon-quark anomalous dimension :math:`\\gamma_{\\gamma d}^{(0,2)}(N)`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    S1 = c.get(c.S1, cache, N)
    tmp = (-16.0 * (-16.0 - 27.0 * N - 13.0 * N**2 - 8.0 * N**3)) / (
        9.0 * (-1.0 + N) * N * (1.0 + N) ** 2
    ) - 16.0 * (2.0 + 3.0 * N + 2.0 * N**2 + N**3) / (
        3.0 * (-1.0 + N) * N * (1.0 + N) ** 2
    ) * S1
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    return constants.ed2 * as1aem1.gamma_phq(N, cache) / constants.CF + eSigma2 * tmp


@nb.njit(cache=True)
def gamma_nspu(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` singlet-like non-singlet anomalous dimension for up quarks.

    Implements sum of Eqs. (57-58) of :cite:`deFlorian:2016gvk` for q=u.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` singlet-like non-singlet anomalous dimension
        :math:`\\gamma_{ns,+,u}^{(0,2)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    tmp = (
        2.0
        * (-12.0 + 20.0 * N + 47.0 * N**2 + 6.0 * N**3 + 3.0 * N**4)
        / (9.0 * N**2 * (1.0 + N) ** 2)
        - 80.0 / 9.0 * S1
        + 16.0 / 3.0 * S2
    ) * eSigma2
    return constants.eu2 * as1aem1.gamma_nsp(N, cache) / constants.CF / 2.0 + tmp


@nb.njit(cache=True)
def gamma_nspd(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` singlet-like non-singlet anomalous dimension for down quarks.

    Implements sum of Eqs. (57-58) of :cite:`deFlorian:2016gvk` for q=d.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` singlet-like non-singlet anomalous dimension
        :math:`\\gamma_{ns,+,d}^{(0,2)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    tmp = (
        2.0
        * (-12.0 + 20.0 * N + 47.0 * N**2 + 6.0 * N**3 + 3.0 * N**4)
        / (9.0 * N**2 * (1.0 + N) ** 2)
        - 80.0 / 9.0 * S1
        + 16.0 / 3.0 * S2
    ) * eSigma2
    return constants.ed2 * as1aem1.gamma_nsp(N, cache) / constants.CF / 2.0 + tmp


@nb.njit(cache=True)
def gamma_nsmu(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` valence-like non-singlet anomalous dimension for up quarks.

    Implements difference between Eqs. (57-58) of :cite:`deFlorian:2016gvk` for q=u.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` valence-like non-singlet anomalous dimension
        :math:`\\gamma_{ns,-,u}^{(0,2)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    tmp = (
        2
        * (-12.0 + 20.0 * N + 47.0 * N**2 + 6.0 * N**3 + 3.0 * N**4)
        / (9.0 * N**2 * (1.0 + N) ** 2)
        - 80.0 / 9.0 * S1
        + 16.0 / 3.0 * S2
    ) * eSigma2
    return constants.eu2 * as1aem1.gamma_nsm(N, cache) / constants.CF / 2.0 + tmp


@nb.njit(cache=True)
def gamma_nsmd(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` valence-like non-singlet anomalous dimension for down quarks.

    Implements difference between Eqs. (57-58) of :cite:`deFlorian:2016gvk` for q=d.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` valence-like non-singlet anomalous dimension
        :math:`\\gamma_{ns,-,d}^{(0,2)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    tmp = (
        2.0
        * (-12.0 + 20.0 * N + 47.0 * N**2 + 6.0 * N**3 + 3.0 * N**4)
        / (9.0 * N**2 * (1.0 + N) ** 2)
        - 80.0 / 9.0 * S1
        + 16.0 / 3.0 * S2
    ) * eSigma2
    return constants.ed2 * as1aem1.gamma_nsm(N, cache) / constants.CF / 2.0 + tmp


@nb.njit(cache=True)
def gamma_ps(N, nf):
    r"""Compute the :math:`O(a_{em}^2)` pure-singlet quark-quark anomalous dimension.

    Implements Eq. (59) of :cite:`deFlorian:2016gvk`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors

    Returns
    -------
    complex
        :math:`O(a_{em}^2)` pure-singlet quark-quark anomalous dimension
        :math:`\\gamma_{ps}^{(0,2)}(N)`

    """
    result = (
        -4.0
        * (2.0 + N * (5.0 + N))
        * (4.0 + N * (4.0 + N * (7.0 + 5.0 * N)))
        / ((-1.0 + N) * N**3 * (1.0 + N) ** 3 * (2.0 + N) ** 2)
    )
    return 2 * nf * constants.CA * result


@nb.njit(cache=True)
def gamma_singlet(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` singlet sector.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        :math:`O(a_{em}^2)` singlet anomalous dimension :math:`\\gamma_{S}^{(0,2)}`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    vu = nu / nf
    vd = nd / nf
    e2m = constants.eu2 - constants.ed2
    e2avg = (nu * constants.eu2 + nd * constants.ed2) / nf
    e2m = constants.eu2 - constants.ed2
    gamma_ph_u = gamma_phu(N, nf, cache)
    gamma_ph_d = gamma_phd(N, nf, cache)
    gamma_u_ph = gamma_uph(N, nf, cache)
    gamma_d_ph = gamma_dph(N, nf, cache)
    gamma_ns_p_u = gamma_nspu(N, nf, cache)
    gamma_ns_p_d = gamma_nspd(N, nf, cache)
    gamma_pure_singlet = gamma_ps(N, nf)
    gamma_S_02 = np.array(
        [
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            [
                0.0 + 0.0j,
                gamma_phph(N, nf),
                vu * constants.eu2 * gamma_ph_u + vd * constants.ed2 * gamma_ph_d,
                vu * (constants.eu2 * gamma_ph_u - constants.ed2 * gamma_ph_d),
            ],
            [
                0.0 + 0.0j,
                vu * constants.eu2 * gamma_u_ph + vd * constants.ed2 * gamma_d_ph,
                vu * constants.eu2 * gamma_ns_p_u
                + vd * constants.ed2 * gamma_ns_p_d
                + e2avg**2 * gamma_pure_singlet,
                vu
                * (
                    constants.eu2 * gamma_ns_p_u
                    - constants.ed2 * gamma_ns_p_d
                    + e2m * e2avg * gamma_pure_singlet
                ),
            ],
            [
                0.0 + 0.0j,
                vd * (constants.eu2 * gamma_u_ph - constants.ed2 * gamma_d_ph),
                vd
                * (
                    constants.eu2 * gamma_ns_p_u
                    - constants.ed2 * gamma_ns_p_d
                    + e2m * e2avg * gamma_pure_singlet
                ),
                vd * constants.eu2 * gamma_ns_p_u
                + vu * constants.ed2 * gamma_ns_p_d
                + vu * vd * e2m**2 * gamma_pure_singlet,
            ],
        ],
        np.complex_,
    )
    return gamma_S_02


@nb.njit(cache=True)
def gamma_valence(N, nf, cache):
    r"""Compute the :math:`O(a_{em}^2)` valence sector.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        :math:`O(a_{em}^2)` valence anomalous dimension :math:`\\gamma_{V}^{(0,2)}`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    vu = nu / nf
    vd = nd / nf
    gamma_ns_m_u = gamma_nsmu(N, nf, cache)
    gamma_ns_m_d = gamma_nsmd(N, nf, cache)
    gamma_V_02 = np.array(
        [
            [
                vu * constants.eu2 * gamma_ns_m_u + vd * constants.ed2 * gamma_ns_m_d,
                vu * (constants.eu2 * gamma_ns_m_u - constants.ed2 * gamma_ns_m_d),
            ],
            [
                vd * (constants.eu2 * gamma_ns_m_u - constants.ed2 * gamma_ns_m_d),
                vd * constants.eu2 * gamma_ns_m_u + vu * constants.ed2 * gamma_ns_m_d,
            ],
        ],
        np.complex_,
    )
    return gamma_V_02



"""The |NLO| Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np

from eko.constants import CA, CF, TR, zeta2, zeta3

from ....harmonics import cache as c

# Non Singlet sector is swapped
from ...unpolarized.space_like.as2 import gamma_nsm as gamma_nsp
from ...unpolarized.space_like.as2 import gamma_nsp as gamma_nsm


@nb.njit(cache=True)
def gamma_ps(n, nf):
    r"""Compute the |NLO| polarized pure-singlet quark-quark anomalous dimension :cite:`Gluck:1995yr` (eq A.3).

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        number of active flavors

    Returns
    -------
    complex
        |NLO| pure-singlet quark-quark anomalous dimension :math:`\\gamma_{ps}^{(1)}(n)`

    """
    gqqps1_nfcf = (2 * (n + 2) * (1 + 2 * n + n**3)) / ((1 + n) ** 3 * n**3)
    result = 4.0 * TR * nf * CF * gqqps1_nfcf
    return result


@nb.njit(cache=True)
def gamma_qg(n, nf, cache):
    r"""Compute the |NLO| polarized quark-gluon singlet anomalous dimension :cite:`Gluck:1995yr` (eq A.4).

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        |NLO| quark-gluon singlet anomalous dimension :math:`\\gamma_{qg}^{(1)}(n)`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    Sp2m = c.get(c.S2mh, cache, n)
    gqg1_nfca = (
        (S1**2 - S2 + Sp2m) * (n - 1) / (n * (n + 1))
        - 4 * S1 / (n * (1 + n) ** 2)
        - (-2 - 7 * n + 3 * n**2 - 4 * n**3 + n**4 + n**5)
        / (n**3 * (1 + n) ** 3)
    ) * 2.0
    gqg1_nfcf = (
        (-(S1**2) + S2 + 2 * S1 / n) * (n - 1) / (n * (n + 1))
        - (n - 1)
        * (1 + 3.5 * n + 4 * n**2 + 5 * n**3 + 2.5 * n**4)
        / (n**3 * (1 + n) ** 3)
        + 4 * (n - 1) / (n**2 * (1 + n) ** 2)
    ) * 2
    result = 4.0 * TR * nf * (CA * gqg1_nfca + CF * gqg1_nfcf)
    return result


@nb.njit(cache=True)
def gamma_gq(n, nf, cache):
    r"""Compute the |NLO| polarized gluon-quark singlet anomalous dimension :cite:`Gluck:1995yr` (eq A.5).

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        |NLO| gluon-quark singlet anomalous dimension :math:`\\gamma_{gq}^{(1)}(n)`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    Sp2m = c.get(c.S2mh, cache, n)
    ggq1_cfcf = (
        (2 * (S1**2 + S2) * (n + 2)) / (n * (n + 1))
        - (2 * S1 * (n + 2) * (1 + 3 * n)) / (n * (1 + n) ** 2)
        - ((n + 2) * (2 + 15 * n + 8 * n**2 - 12.0 * n**3 - 9.0 * n**4))
        / (n**3 * (1 + n) ** 3)
        + 8 * (n + 2) / (n**2 * (1 + n) ** 2)
    ) * 0.5
    ggq1_cfca = -(
        -(-(S1**2) - S2 + Sp2m) * (n + 2) / (n * (n + 1))
        - S1 * (12 + 22 * n + 11 * n**2) / (3 * n**2 * (n + 1))
        + (36 + 72 * n + 41 * n**2 + 254 * n**3 + 271 * n**4 + 76 * n**5)
        / (9 * n**3 * (1 + n) ** 3)
    )
    ggq1_cfnf = (-S1 * (n + 2)) / (3 * n * (n + 1)) + ((n + 2) * (2 + 5 * n)) / (
        9 * n * (1 + n) ** 2
    )
    result = 4 * CF * (CA * ggq1_cfca + CF * ggq1_cfcf + 4.0 * TR * nf * ggq1_cfnf)
    return result


@nb.njit(cache=True)
def gamma_gg(n, nf, cache):
    r"""Compute the |NLO| polarized gluon-gluon singlet anomalous dimension :cite:`Gluck:1995yr` (eq A.6).

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        |NLO| gluon-quark singlet anomalous dimension :math:`\\gamma_{gq}^{(1)}(n)`

    """
    S1 = c.get(c.S1, cache, n)
    Sp1m = c.get(c.S1mh, cache, n)
    Sp2m = c.get(c.S2mh, cache, n)
    Sp3m = c.get(c.S3mh, cache, n)
    S1h = c.get(c.S1h, cache, n)
    g3 = c.get(c.g3, cache, n)
    SSCHLM = zeta2 / 2 * (+Sp1m - S1h + 2 / n) - S1 / n**2 - g3 - 5 * zeta3 / 8
    ggg1_caca = (
        -4 * S1 * Sp2m
        - Sp3m
        + 8 * SSCHLM
        + 8 * Sp2m / (n * (n + 1))
        + 2.0
        * S1
        * (72 + 144 * n + 67 * n**2 + 134 * n**3 + 67 * n**4)
        / (9 * n**2 * (n + 1) ** 2)
        - (
            144
            + 258 * n
            + 7 * n**2
            + 698 * n**3
            + 469 * n**4
            + 144 * n**5
            + 48 * n**6
        )
        / (9 * n**3 * (1 + n) ** 3)
    ) * 0.5
    ggg1_canf = (
        -5 * S1 / 9
        + (-3 + 13 * n + 16 * n**2 + 6 * n**3 + 3 * n**4)
        / (9 * n**2 * (1 + n) ** 2)
    ) * 4
    ggg1_cfnf = (4 + 2 * n - 8 * n**2 + n**3 + 5 * n**4 + 3 * n**5 + n**6) / (
        n**3 * (1 + n) ** 3
    )
    result = 4 * (CA**2 * ggg1_caca + TR * nf * (CA * ggg1_canf + CF * ggg1_cfnf))
    return result


@nb.njit(cache=True)
def gamma_singlet(n, nf, cache):
    r"""Compute the |NLO| polarized singlet anomalous dimension matrix.

        .. math::
            \gamma_S^{(1)} = \left(\begin{array}{cc}
            \gamma_{qq}^{(1)} & \gamma_{qg}^{(1)}\\
            \gamma_{gq}^{(1)} & \gamma_{gg}^{(1)}
            \end{array}\right)

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        |NLO| singlet anomalous dimension matrix :math:`\gamma_{S}^{(1)}(N)`

    """
    gamma_qq = gamma_nsp(n, nf, cache) + gamma_ps(n, nf)
    gamma_S_0 = np.array(
        [
            [gamma_qq, gamma_qg(n, nf, cache)],
            [gamma_gq(n, nf, cache), gamma_gg(n, nf, cache)],
        ],
        np.complex_,
    )
    return gamma_S_0



r"""Contains the scale variation for ``ModSV=exponentiated``."""
import numba as nb

from .. import beta


@nb.njit(cache=True)
def gamma_variation(gamma, order, nf, L):
    """Adjust the anomalous dimensions with the scale variations.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    order : tuple(int,int)
        perturbation order
    nf : int
        number of active flavors
    L : float
        logarithmic ratio of factorization and renormalization scale

    Returns
    -------
    numpy.ndarray
        adjusted anomalous dimensions
    """
    # since we are modifying *in-place* be careful, that the order matters!
    # and indeed, we need to adjust the highest elements first
    beta0 = beta.beta_qcd((2, 0), nf)
    beta1 = beta.beta_qcd((3, 0), nf)
    beta2 = beta.beta_qcd((4, 0), nf)
    if order[0] >= 4:
        gamma[3] += (
            3.0 * beta0 * L * gamma[2]
            + (2.0 * beta1 * L + 3.0 * beta0**2 * L**2) * gamma[1]
            + (beta2 * L + 5.0 / 2.0 * beta1 * beta0 * L**2 + beta0**3 * L**3)
            * gamma[0]
        )
    if order[0] >= 3:
        gamma[2] += (
            2.0 * beta0 * gamma[1] * L + (beta1 * L + beta0**2 * L**2) * gamma[0]
        )
    if order[0] >= 2:
        gamma[1] += beta0 * gamma[0] * L
    return gamma


@nb.njit(cache=True)
def gamma_variation_qed(gamma, order, nf, L, alphaem_running):
    """Adjust the anomalous dimensions with the scale variations.

    Parameters
    ----------
    gamma : numpy.ndarray
        anomalous dimensions
    order : tuple(int,int)
        perturbation order
    nf : int
        number of active flavors
    L : float
        logarithmic ratio of factorization and renormalization scale

    Returns
    -------
    gamma : numpy.ndarray
        adjusted anomalous dimensions
    """
    # if alphaem is fixed then only alphas is varied so gamma[0,1] and gamma[0,2]
    # don't get a variation while gamma[1,1] gets a variation that is O(as2aem1)
    # that we are neglecting
    gamma[1:, 0] = gamma_variation(gamma[1:, 0], order, nf, L)
    if alphaem_running:
        if order[1] >= 2:
            beta0qed = beta.beta_qed((0, 2), nf)
            gamma[0, 2] += beta0qed * gamma[0, 1] * L
        return gamma



"""Library providing all necessary tools for PDF interpolation.

This library also provides a class to generate the interpolator :class:`InterpolatorDispatcher`.
Upon construction the dispatcher generates a number of functions
to evaluate the interpolator.
"""
import logging
import math
from typing import Sequence, Union

import numba as nb
import numpy as np
import numpy.typing as npt
import scipy.special as sp

logger = logging.getLogger(__name__)


class Area:
    """Define area of subgrid interpolators.

    Upon construction an array of coefficients is generated.

    Parameters
    ----------
    lower_index: int
        lower index of the area
    poly_number: int
        number of polynomial
    block: tuple(int, int)
        kmin and kmax
    xgrid: numpy.ndarray
        Grid in x-space from which the interpolators are constructed

    """

    def __init__(self, lower_index, poly_number, block, xgrid):
        # check range
        if poly_number < block[0] or block[1] < poly_number:
            raise ValueError(
                f"polynom #{poly_number} cannot be a part of the block which"
                f"spans from {block[0]} to {block[1]}"
            )
        self.xmin = xgrid[lower_index]
        self.xmax = xgrid[lower_index + 1]
        self.poly_number = poly_number
        self.kmin = block[0]
        self.kmax = block[1]
        self.coefs = self._compute_coefs(xgrid)

    def _reference_indices(self):
        """Iterate over all indices which are part of the block."""
        for k in range(self.kmin, self.kmax + 1):
            if k != self.poly_number:
                yield k

    def _compute_coefs(self, xgrid):
        """Compute the coefficients for this area given a grid on :math:`x`."""
        denominator = 1.0
        coeffs = np.ones(1)
        xj = xgrid[self.poly_number]
        for s, k in enumerate(self._reference_indices()):
            xk = xgrid[k]
            denominator *= xj - xk
            Mxk_coeffs = -xk * coeffs
            coeffs = np.concatenate(([0.0], coeffs))
            coeffs[: s + +1] += Mxk_coeffs
        coeffs /= denominator
        return coeffs

    def __iter__(self):
        """Iterate the generated coefficients."""
        yield from self.coefs


@nb.njit(cache=True)
def log_evaluate_Nx(N, logx, area_list):
    r"""Evaluate logarithmic interpolator in N-space.

    A single logarithmic Lagrange interpolator is evaluated in N-space
    multiplied by the Mellin-inversion factor.

    .. math::
        \tilde p(N)*\exp(- N * \ln(x))

    Parameters
    ----------
    N : complex
        Mellin variable
    logx : float
        logarithm of inversion point
    area_list : list
        area configuration of basis function

    Returns
    -------
    res : float
        kernel * inversion factor

    """
    res = 0.0
    for a in area_list:
        logxmin = a[0]
        logxmax = a[1]
        coefs = a[2:]
        # skip area completely?
        if logx >= logxmax or np.abs(logx - logxmax) < _atol_eps:
            continue
        umax = N * logxmax
        umin = N * logxmin
        emax = np.exp(N * (logxmax - logx))
        emin = np.exp(N * (logxmin - logx))
        for i, coef in enumerate(coefs):
            tmp = 0.0
            facti = math.gamma(i + 1) * pow(-1, i) / pow(N, i + 1)
            for k in range(i + 1):
                factk = 1.0 / math.gamma(k + 1)
                # this condition is actually not necessary in python since
                # there pow(0,0) == 1 and apparently this is inherited in
                # Numba/C, however this is mathematically cleaner
                if np.abs(umax) < _atol_eps and k == 0:
                    pmax = emax
                else:
                    pmax = pow(-umax, k) * emax
                # drop factor by analytics?
                if logx >= logxmin or np.abs(logx - logxmin) < _atol_eps:
                    pmin = 0
                else:
                    pmin = pow(-umin, k) * emin
                tmp += factk * (pmax - pmin)
            res += coef * facti * tmp
    return res


@nb.njit(cache=True)
def evaluate_Nx(N, logx, area_list):
    r"""Evaluate linear interpolator in N-space.

    A single linear Lagrange interpolator is evaluated in N-space multiplied by
    the Mellin-inversion factor.

    .. math::
        \tilde p(N)*\exp(- N * \ln(x))

    Parameters
    ----------
    N : complex
        Mellin variable
    logx : float
        logarithm of inversion point
    area_list : list
        area configuration of basis function

    Returns
    -------
    res : float
        basis function * inversion factor

    """
    res = 0.0
    for a in area_list:
        xmin = a[0]
        xmax = a[1]
        coefs = a[2:]
        lnxmax = np.log(xmax)
        # skip area completely?
        if logx >= lnxmax:
            continue
        for i, coef in enumerate(coefs):
            if xmin == 0.0:
                low = 0.0
            else:
                lnxmin = np.log(xmin)
                low = np.exp(N * (lnxmin - logx) + i * lnxmin)
            up = np.exp(N * (lnxmax - logx) + i * lnxmax)
            res += coef * (up - low) / (N + i)
    return res


@nb.njit(cache=True)
def evaluate_grid(N, is_log, logx, area_list):
    """Evaluate interpolator in N-space.

    Parameters
    ----------
    N : complex
        Mellin variable
    is_log : boolean
        is a logarithmic interpolation
    logx : float
        logarithm of inversion point
    area_list : list
        area configuration of basis function

    Returns
    -------
    pj : float
        basis function * inversion factor

    """
    if is_log:
        pj = log_evaluate_Nx(N, logx, area_list)
    else:
        pj = evaluate_Nx(N, logx, area_list)
    return pj


# TODO lift to runcard?
_atol_eps = 10 * np.finfo(float).eps


@nb.njit(cache=True)
def evaluate_x(x, area_list):
    """Evaluate a single linear Lagrange interpolator in x-space.

    .. math::
        p(x)

    Parameters
    ----------
        x : float
            interpolation point
        area_list : list
            area configuration of basis function

    Returns
    -------
        res : float
            basis function(x)
    """
    res = 0.0
    for j, a in enumerate(area_list):
        xmin = a[0]
        xmax = a[1]
        coefs = a[2:]
        if xmin < x <= xmax or (j == 0 and np.abs(x - xmin) < _atol_eps):
            for i, coef in enumerate(coefs):
                res += coef * pow(x, i)
            return res

    return res


@nb.njit(cache=True)
def log_evaluate_x(x, area_list):
    """Evaluate a single logarithmic Lagrange interpolator in x-space.

    .. math::
        p(x)

    Parameters
    ----------
    x : float
        interpolation point
    area_list : list
        area configuration of basis function

    Returns
    -------
    res : float
        basis function(x)

    """
    x = np.log(x)
    return evaluate_x(x, area_list)


class BasisFunction:
    """Represent an element of the polynomial basis.

    It contains a list of areas for a given polynomial number defined by
    (xmin-xmax) which in turn contain a list of coefficients.

    Upon construction it will generate all areas and generate and compile a
    function to evaluate in N (or x) the interpolator

    Parameters
    ----------
    xgrid : numpy.ndarray
        Grid in x-space from which the interpolators are constructed
    poly_number : int
        number of polynomial
    list_of_blocks: list(tuple(int, int))
        list of tuples with the (kmin, kmax) values for each area
    mode_log: bool
        use logarithmic interpolation?
    mode_N: bool
        if true compiles the function on N, otherwise compiles on x

    """

    def __init__(
        self,
        xgrid,
        poly_number,
        list_of_blocks,
        mode_log=True,
        mode_N=True,
    ):
        self.poly_number = poly_number
        self.areas = []
        self._mode_log = mode_log
        self.mode_N = mode_N

        # create areas
        for i, block in enumerate(list_of_blocks):
            if block[0] <= poly_number <= block[1]:
                new_area = Area(i, self.poly_number, block, xgrid)
                self.areas.append(new_area)
        if not self.areas:
            raise ValueError("Error: no areas were generated")
        self.areas_representation = self.areas_to_const()

        # compile
        # TODO move this to InterpolatorDispatcher
        self.callable = None
        if self.mode_N:
            self.compile_n()
        else:
            self.compile_x()

    def is_below_x(self, x):
        """Check all areas to be below specified threshold.

        Parameters
        ----------
        x : float
            reference value

        Returns
        -------
        is_below_x : bool
            xmax of highest area <= x?

        """
        # Log if needed
        if self._mode_log:
            x = np.log(x)
        # note that ordering is important!
        return self.areas[-1].xmax <= x

    def areas_to_const(self):
        """Return an array containing all areas.

        The area format is (`xmin`, `xmax`, `numpy.array` of coefficients).

        Returns
        -------
        numpy.ndarray
            area config

        """
        # This is necessary as numba will ask for everything
        # to be immutable
        area_list = []
        for area in self:
            area_list.append([area.xmin, area.xmax, *area.coefs])
        return np.array(area_list)

    def compile_x(self):
        """Compile the function to evaluate the interpolator in x space."""
        if self._mode_log:
            self.callable = log_evaluate_x
        else:
            self.callable = evaluate_x

    def evaluate_x(self, x):
        """Evaluate basis function in x-space (regardless of the true space).

        Parameters
        ----------
        x : float
            evaluated point

        Returns
        -------
        res : float
            p(x)

        """
        if self.mode_N:
            old_call = self.callable
            self.compile_x()
            res = self.callable(x, self.areas_representation)
            self.callable = old_call
        else:
            res = self.callable(x, self.areas_representation)
        return res

    def compile_n(self):
        r"""Compile the function to evaluate the interpolator in N space.

        Generates a function :meth:`evaluate_Nx` with a (N, logx) signature.

        .. math::
            \tilde p(N)*\exp(- N * \ln(x))

        The polynomials contain naturally factors of :math:`\exp(N * j * \ln(x_{min/max}))`
        which can be joined with the Mellin inversion factor.

        """
        if self._mode_log:
            self.callable = log_evaluate_Nx
        else:
            self.callable = evaluate_Nx

    def __iter__(self):
        yield from self.areas

    def __call__(self, *args, **kwargs):
        """Evaluate function."""
        args = list(args)
        args.append(self.areas_representation)
        return self.callable(*args, **kwargs)


class XGrid:
    """Grid of points in :math:`x`-space.

    This object represents a suitable grid of momentum fractions to be used to
    evaluate the PDF over.

    """

    def __init__(self, xgrid: Union[Sequence, npt.NDArray], log: bool = True):
        ugrid = np.array(np.unique(xgrid), np.float_)
        if len(xgrid) != len(ugrid):
            raise ValueError(f"xgrid is not unique: {xgrid}")
        if len(xgrid) < 2:
            raise ValueError(f"xgrid needs at least 2 points, received {len(xgrid)}")

        self.log = log

        # henceforth ugrid might no longer be the input!
        # which is ok, because for most of the code this is all we need to do
        # to distinguish log and non-log
        if log:
            self._raw = ugrid
            ugrid = np.log(ugrid)

        self.grid = ugrid

    def __len__(self) -> int:
        return len(self.grid)

    def __eq__(self, other) -> bool:
        """Check equality."""
        # check shape before comparing values
        return len(self) == len(other) and np.allclose(self.raw, other.raw)

    @property
    def raw(self) -> np.ndarray:
        """Untransformed grid."""
        return self.grid if not self.log else self._raw

    @property
    def size(self) -> int:
        """Number of pointrs."""
        return self.grid.size

    def tolist(self) -> list:
        """Raw grid as Python list."""
        return self.raw.tolist()

    def dump(self) -> dict:
        """Representation as dictionary."""
        return dict(grid=self.tolist(), log=self.log)

    @classmethod
    def load(cls, doc: dict):
        """Create object from dictinary."""
        return cls(doc["grid"], log=doc["log"])

    @classmethod
    def fromcard(cls, value: list, log: bool):
        """Create object from theory card config.

        The config can invoke other grid generation methods.

        """
        if len(value) == 0:
            raise ValueError("Empty xgrid!")

        if value[0] == "make_grid":
            xgrid = make_grid(*value[1:])
        elif value[0] == "lambertgrid":
            xgrid = lambertgrid(*value[1:])
        else:
            xgrid = np.array(value)

        return cls(xgrid, log=log)


class InterpolatorDispatcher:
    """Setup the interpolator.

    Upon construction will generate a list of :class:`BasisFunction` objects.
    Each of these :class:`BasisFunction` objects expose a `callable`
    method (also accessible as the `__call__` method of the class)
    which will be numba-compiled.


    Parameters
    ----------
    xgrid : numpy.ndarray
        Grid in x-space from which the interpolators are constructed
    polynomial_degree : int
        degree of the interpolation polynomial
    log: bool
        Whether it is a log or linear interpolator
    mode_N: bool
        if true compiles the function on N, otherwise compiles x

    """

    def __init__(
        self, xgrid: Union[XGrid, Sequence, npt.NDArray], polynomial_degree, mode_N=True
    ):
        if not isinstance(xgrid, XGrid):
            xgrid = XGrid(xgrid)

        # sanity checks
        if polynomial_degree < 1:
            raise ValueError(
                f"need at least polynomial_degree 1, received {polynomial_degree}"
            )
        if len(xgrid) <= polynomial_degree:
            raise ValueError(
                f"to interpolate with degree {polynomial_degree} "
                " we need at least that much points + 1"
            )

        # Save the different variables
        self.xgrid = xgrid
        self.polynomial_degree = polynomial_degree
        self.log = xgrid.log
        logger.info(
            "Interpolation: number of points = %d, polynomial degree = %d, logarithmic = %s",
            len(xgrid),
            polynomial_degree,
            xgrid.log,
        )

        # Create blocks
        list_of_blocks = []
        po2 = polynomial_degree // 2
        # if degree is even, we can not split the block symmetric, e.g. deg=2 -> |-|-|
        # so, in case of doubt use the block, which lays higher, i.e.
        # we're not allowed to go so deep -> make po2 smaller
        if polynomial_degree % 2 == 0:
            po2 -= 1
        # iterate areas: there is 1 less then number of points
        for i in range(len(xgrid) - 1):
            kmin = max(0, i - po2)
            kmax = kmin + polynomial_degree
            if kmax >= len(xgrid):
                kmax = len(xgrid) - 1
                kmin = kmax - polynomial_degree
            b = (kmin, kmax)
            list_of_blocks.append(b)

        # Generate the basis functions
        basis_functions = []
        for i in range(len(xgrid)):
            new_basis = BasisFunction(
                xgrid.grid, i, list_of_blocks, mode_log=xgrid.log, mode_N=mode_N
            )
            basis_functions.append(new_basis)
        self.basis = basis_functions

    def __eq__(self, other):
        """Check equality."""
        checks = [
            self.log == other.log,
            self.polynomial_degree == other.polynomial_degree,
            self.xgrid == other.xgrid,
        ]
        return all(checks)

    def __iter__(self):
        # return iter(self.basis)
        yield from self.basis

    def __getitem__(self, item):
        return self.basis[item]

    def get_interpolation(self, targetgrid: Union[npt.NDArray, Sequence]):
        r"""Compute interpolation matrix between `targetgrid` and `xgrid`.

        .. math::
            f(targetgrid) = R \cdot f(xgrid)

        Parameters
        ----------
        targetgrid : array
            grid to interpolate to

        Returns
        -------
        R : array
            interpolation matrix $R_{ij}$, where $i$ is the index over
            `targetgrid`, and $j$ is the index on the internal basis (the
            one stored in the :class:`InterpolatorDispatcher` instance)

        """
        # trivial?
        if len(targetgrid) == len(self.xgrid) and np.allclose(
            targetgrid, self.xgrid.raw
        ):
            return np.eye(len(self.xgrid))
        # compute map
        out = []
        for x in targetgrid:
            row = []
            for b in self.basis:
                row.append(b.evaluate_x(x))
            out.append(row)
        return np.array(out)

    def to_dict(self):
        """Return the configuration for the underlying xgrid.

        An instance can be constructed again with just the information returned
        by this method.

        Note
        ----
        The output dictionary contains a numpy array.

        Returns
        -------
        ret : dict
            full grid configuration

        """
        ret = {
            "xgrid": self.xgrid.dump(),
            "polynomial_degree": self.polynomial_degree,
            "is_log": self.log,
        }
        return ret


def make_grid(
    n_low, n_mid, n_high=0, x_min=1e-7, x_low=0.1, x_high=0.9, x_high_max=1.0 - 1e-4
):
    """Create a log-lin-log-spaced grid.

    1.0 is always part of the grid and the final grid is unified, i.e. esp. points that might
    appear twice at the borders of the regions are unified.

    Parameters
    ----------
    n_low : int
        points in the small-x region
    n_mid : int
        points in the medium-x region
    n_high : int
        points in the large-x region
    x_min : float
        minimum x (included)
    x_low : float
        seperation point between small and medium
    x_high_max : float
        closest point before 1

    Returns
    -------
    xgrid : numpy.ndarray
        generated grid

    """
    # low
    if n_mid == 0:
        x_low = 1.0
    xgrid_low = np.geomspace(x_min, x_low, n_low)
    # high
    if n_high == 0:
        x_high = 1.0
        xgrid_high = np.array([])
    else:
        xgrid_high = 1.0 - np.geomspace(1 - x_high, 1 - x_high_max, n_high)
    # mid
    xgrid_mid = np.linspace(x_low, x_high, n_mid)
    # join
    xgrid = np.unique(np.concatenate((xgrid_low, xgrid_mid, xgrid_high, np.array([1]))))
    return xgrid


def lambertgrid(n_pts, x_min=1e-7, x_max=1.0):
    r"""Create a smoothly spaced grid that is linear near 1 and logarithmic near 0.

    It is generated by the relation:

    .. math::
        x(y) = \frac{1}{5} W_{0}(5 \exp(5-y))

    where :math:`W_{0}` is the principle branch of the
    :func:`Lambert W function <scipy.special.lambertw>` and
    :math:`y` is a variable which extremes are given as function of :math:`x`
    by the direct relation:

    .. math::
        y(x) = 5(1-x)-\log(x)

    This method is implemented in `PineAPPL`, :cite:`Carrazza_2020` eq 2.11 and relative
    paragraph.

    Parameters
    ----------
        n_pts : int
            number of points
        x_min : float
            minimum x (included)
        x_max : float
            maximum x (included)

    Returns
    -------
        xgrid : numpy.ndarray
            generated grid
    """

    def direct_relation(x):
        return 5 * (1 - x) - np.log(x)

    def inverse_relation(y):
        return np.real(1 / 5 * sp.lambertw(5 * np.exp(5 - y)))

    y_min = direct_relation(x_min)
    y_max = direct_relation(x_max)

    return np.array([inverse_relation(y) for y in np.linspace(y_min, y_max, n_pts)])



"""Abstraction for serializations.

A few known types are directly registered here, in order to be transparently
codified in more native structures.

"""
import copy
import dataclasses
import enum
import inspect
import typing

import numpy as np

from .. import interpolation

# TODO: use typing.dataclass_transform, new in Python 3.11, since all child
# classes are supposed to be dataclasses


@dataclasses.dataclass
class DictLike:
    """Dictionary compatibility base class, for dataclasses.

    This class add compatibility to import and export from Python :class:`dict`,
    in such a way to support serialization interfaces working with them.

    Some collections and scalar objects are normalized to native Python
    structures, in order to simplify the on-disk representation.

    """

    @classmethod
    def _from_dict(cls, dictionary):
        """Initialize dataclass object from raw dictionary.

        Parameters
        ----------
        dictionary : dict
            the dictionary to be converted to :class:`DictLike`

        Returns
        -------
        DictLike
            instance with `dictionary` content loaded as attributes

        """
        dictionary = copy.deepcopy(dictionary)

        # allow initialization from tuple, useful for simple dataclasses
        # NOTE: no further nesting will be supported, they are supposed to be
        # simple
        # TODO: update some names consistently, `DictLike` nad `from_dict` are
        # not any longer the best. And it is good in any case to drop the old
        # `from_dict` name, and solve part of the ambiguity with existing
        # methods
        if any(isinstance(dictionary, cls) for cls in [tuple, list]):
            return cls(*dictionary)

        # support nesting, modifying the copied dictionary
        for field in dataclasses.fields(cls):
            # dictionary does not provide any type information we can rely on
            # to load better, so we will assume that it is fully loaded inside
            if field.type in [dict, typing.Dict]:
                continue
            # allow for default values: is sufficient to leave it empty, the
            # default is already in the class constructor
            if (
                field.name not in dictionary
                and field.default is not dataclasses.MISSING
            ):
                continue

            # load nested: the field has a definite type, and we will use this
            # information to automatically implement the loader
            dictionary[field.name] = load_field(field.type, dictionary[field.name])

        # finally construct the class, just passing the arguments by name
        return cls(**dictionary)

    @classmethod
    def from_dict(cls, dictionary):
        """Deserialize, overwritable interface.

        The default implementation is just :meth:`DictLike._from_dict`, but it
        can be safely overwritten (usually transforming the input before a call
        to :meth:`DictLike._from_dict` itself).

        """
        return cls._from_dict(dictionary)

    def _raw(self):
        """Convert dataclass object to raw dictionary.

        Normalize:

            - :class:`np.ndarray` to lists (possibly nested)
            - scalars to the corresponding built-in type (e.g. :class:`float`)
            - :class:`tuple` to lists
            - :class:`interpolation.XGrid` to the intrinsic serialization format

        Returns
        -------
        dict
            dictionary representation

        """
        dictionary = {}

        for field in dataclasses.fields(self):
            dictionary[field.name] = raw_field(getattr(self, field.name))

        return dictionary

    @property
    def raw(self):
        """Overwritable serialization.

        The default implementation is just :meth:`DictLike._raw`, but it can be
        safely overwritten (usually starting from :meth:`DictLike._raw`
        itself).

        """
        return self._raw()

    @property
    def public_raw(self):
        """Serialize only public attributes."""
        return {k: v for k, v in self._raw().items() if not k.startswith("_")}


def load_field(type_, value):
    """Deserialize dataclass field."""
    # TODO: nice place for a match statement...
    if type(type_) is typing.NewType:
        return load_field(type_.__supertype__, value)
    try:
        # before py3.10 typing.NewType was just a function, so the check above
        # would  fail
        return load_field(type_.__supertype__, value)
    except AttributeError:
        pass

    if typing.get_origin(type_) is not None:
        # this has to go first since for followin ones I will assume they are
        # valid classes, cf. the module docstring
        return load_typing(type_, value)

    # FIXME: python 3.8 does not recognize the origin of `npt.NDArray`, but at
    # the same time it is not a proper class since `numpy==1.23.2`, since it is
    # not recognized by `inspect.isclass`, and `type(npt.NDArray)` is not
    # inheriting from `type`
    # However, this check should be matched by every version of NumPy for every
    # version of Python supported
    # Once py3.8 will be dropped, move this back below `inspect.isclass`
    # assertion
    if np.ndarray in type_.__mro__ or issubclass(type_, np.ndarray):
        # do not apply array on scalars
        if isinstance(value, list):
            return np.array(value)
        return value

    assert inspect.isclass(type_)

    if issubclass(type_, DictLike):
        return type_.from_dict(value)
    if issubclass(type_, enum.Enum):
        return load_enum(type_, value)
    if isinstance(value, dict):
        return type_(**value)

    return type_(value)


def load_enum(type_, value):
    """Deserialize enum variant.

    Accepts both the name and value of variants, attempted in this order.

    Raises
    ------
    ValueError
        if `value` is not the name nor the value of any enum variant

    """
    try:
        return type_[value]
    except KeyError:
        return type_(value)


def load_typing(type_, value):
    """Deserialize type hint associated field."""
    origin = typing.get_origin(type_)
    assert origin is not None

    # unions have to be explored over their variants
    if origin is typing.Union:
        for variant in typing.get_args(type_):
            try:
                return load_field(variant, value)
            except (TypeError, ValueError):
                ...

        if type(None) in type_.__args__:
            return None
        raise TypeError

    if issubclass(origin, (list, typing.Generic)):
        T = typing.get_args(type_)[0]
        return origin([load_field(T, x) for x in value])

    return load_field(origin, value)


def raw_field(value):
    """Serialize DictLike field."""
    # replace numpy arrays with lists
    if isinstance(value, np.ndarray):
        return value.tolist()
    # replace numpy scalars with python ones
    if isinstance(value, float):
        return float(value)
    if isinstance(value, interpolation.XGrid):
        return value.dump()["grid"]
    if isinstance(value, tuple):
        return list(value)
    if isinstance(value, enum.Enum):
        return value.value
    if isinstance(value, DictLike):
        return value.raw
    if dataclasses.is_dataclass(value):
        # not supporting nested DictLike inside nested plain dataclasses
        return dataclasses.asdict(value)
    if isinstance(value, list):
        return [raw_field(el) for el in value]

    return value



"""Apply operator evolution to PDF set."""

import numpy as np

from eko import basis_rotation as br
from eko import interpolation
from eko.io import EKO


def apply_pdf(
    eko: EKO,
    lhapdf_like,
    targetgrid=None,
    rotate_to_evolution_basis=False,
    qed=False,
):
    """
    Apply all available operators to the input PDFs.

    Parameters
    ----------
        output : eko.output.EKO
            eko output object containing all informations
        lhapdf_like : object
            object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
            and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
        targetgrid : list
            if given, interpolates to the pdfs given at targetgrid (instead of xgrid)
        rotate_to_evolution_basis : bool
            if True rotate to evoluton basis

    Returns
    -------
        out_grid : dict
            output PDFs and their associated errors for the computed mu2grid
    """
    if rotate_to_evolution_basis:
        if not qed:
            rotate_flavor_to_evolution = br.rotate_flavor_to_evolution
        else:
            rotate_flavor_to_evolution = br.rotate_flavor_to_unified_evolution
        return apply_pdf_flavor(
            eko, lhapdf_like, targetgrid, rotate_flavor_to_evolution, qed
        )
    return apply_pdf_flavor(eko, lhapdf_like, targetgrid)


CONTRACTION = "ajbk,bk"


def apply_pdf_flavor(
    eko: EKO, lhapdf_like, targetgrid=None, flavor_rotation=None, qed=False
):
    """
    Apply all available operators to the input PDFs.

    Parameters
    ----------
        output : eko.output.EKO
            eko output object containing all informations
        lhapdf_like : object
            object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
            and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
        targetgrid : list
            if given, interpolates to the pdfs given at targetgrid (instead of xgrid)
        flavor_rotation : np.ndarray
            Rotation matrix in flavor space
        qed : bool
            activate qed

    Returns
    -------
        out_grid : dict
            output PDFs and their associated errors for the computed mu2grid
    """
    # create pdfs
    pdfs = np.zeros((len(eko.bases.inputpids), len(eko.bases.inputgrid)))
    for j, pid in enumerate(eko.bases.inputpids):
        if not lhapdf_like.hasFlavor(pid):
            continue
        pdfs[j] = np.array(
            [lhapdf_like.xfxQ2(pid, x, eko.mu20) / x for x in eko.bases.inputgrid.raw]
        )

    # build output
    out_grid = {}
    for (mu2, nf), elem in eko.items():
        pdf_final = np.einsum(CONTRACTION, elem.operator, pdfs, optimize="optimal")
        if elem.error is not None:
            error_final = np.einsum(CONTRACTION, elem.error, pdfs, optimize="optimal")
        else:
            error_final = None
        out_grid[mu2] = {
            "pdfs": dict(zip(eko.bases.targetpids, pdf_final)),
            "errors": None,
        }
        if error_final is not None:
            out_grid[mu2]["errors"] = dict(zip(eko.bases.targetpids, error_final))

    # rotate to evolution basis
    if flavor_rotation is not None:
        for q2, op in out_grid.items():
            pdf = flavor_rotation @ np.array(
                [op["pdfs"][pid] for pid in br.flavor_basis_pids]
            )
            if not qed:
                evol_basis = br.evol_basis
            else:
                evol_basis = br.unified_evol_basis
            op["pdfs"] = dict(zip(evol_basis, pdf))
            if op["errors"] is not None:
                errors = flavor_rotation @ np.array(
                    [op["errors"][pid] for pid in br.flavor_basis_pids]
                )
                op["errors"] = dict(zip(evol_basis, errors))

    # rotate/interpolate to target grid
    if targetgrid is not None:
        b = interpolation.InterpolatorDispatcher(
            xgrid=eko.bases.targetgrid,
            polynomial_degree=eko.operator_card.configs.interpolation_polynomial_degree,
            mode_N=False,
        )

        rot = b.get_interpolation(targetgrid)
        for evpdf in out_grid.values():
            for pdf_label in evpdf["pdfs"]:
                evpdf["pdfs"][pdf_label] = np.matmul(rot, evpdf["pdfs"][pdf_label])
                if evpdf["errors"] is not None:
                    evpdf["errors"][pdf_label] = np.matmul(
                        rot, evpdf["errors"][pdf_label]
                    )

    return out_grid



"""Generic utilities to work with EKOs."""
import os
from typing import Optional

import numpy as np

from eko.io.struct import EKO, Operator

CONTRACTION = "ajbk,bkcl -> ajcl"


def ekos_product(
    eko_ini: EKO,
    eko_fin: EKO,
    rtol: float = 1e-6,
    atol: float = 1e-10,
    path: Optional[os.PathLike] = None,
):
    """Compute the product of two ekos.

    Parameters
    ----------
    eko_ini :
        initial eko operator
    eko_fin :
        final eko operator
    rtol :
        relative tolerance on Q2, used to check compatibility
    atol :
        absolute tolerance on Q2, used to check compatibility
    path :
        if not provided, the operation is done in-place, otherwie a new
        operator is written at the given path

    """
    # TODO: add a control on the theory (but before we need to implement
    # another kind of output which includes the theory and operator runcards)

    ep_match = eko_ini.approx(
        (eko_fin.operator_card.mu0**2, eko_fin.theory_card.heavy.num_flavs_init),
        rtol=rtol,
        atol=atol,
    )
    if ep_match is None:
        raise ValueError(
            "Initial Q2 of final eko operator does not match any final Q2 in"
            " the initial eko operator"
        )
    ope1 = eko_ini[ep_match]
    assert ope1 is not None

    ope1_value = ope1.operator.copy()
    ope1_error = ope1.error
    if ope1_error is not None:
        ope1_error = ope1_error.copy()

    if path is None:
        final_eko = eko_ini
    else:
        eko_ini.deepcopy(path)
        final_eko = EKO.edit(path)

    for q2, op2 in eko_fin.items():
        if q2 in eko_ini:
            continue

        op = np.einsum(CONTRACTION, ope1_value, op2.operator)

        if ope1_error is not None and op2.error is not None:
            error = np.einsum(CONTRACTION, ope1_value, op2.error) + np.einsum(
                CONTRACTION, ope1_error, op2.operator
            )
        else:
            error = None

        final_eko[q2] = Operator(operator=op, error=error)

    if path is not None:
        final_eko.close()



"""The :math:`O(a_s^1a_{em}^1)` Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np

from eko import constants
from eko.constants import zeta2, zeta3

from ....harmonics import cache as c


@nb.njit(cache=True)
def gamma_phq(N, cache):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` photon-quark anomalous dimension.

    Implements Eq. (36) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache


    Returns
    -------
    complex
        :math:`O(a_s^1a_{em}^1)` photon-quark anomalous dimension :math:`\\gamma_{\\gamma q}^{(1,1)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    tmp_const = (
        2.0
        * (
            -4.0
            - 12.0 * N
            - N**2
            + 28.0 * N**3
            + 43.0 * N**4
            + 30.0 * N**5
            + 12.0 * N**6
        )
        / ((-1.0 + N) * N**3 * (1.0 + N) ** 3)
    )
    tmp_S1 = (
        -4.0
        * (10.0 + 27.0 * N + 25.0 * N**2 + 13.0 * N**3 + 5.0 * N**4)
        / ((-1.0 + N) * N * (1.0 + N) ** 3)
    )
    tmp_S12 = 4.0 * (2.0 + N + N**2) / ((-1.0 + N) * N * (1.0 + N))
    tmp_S2 = 4.0 * (2.0 + N + N**2) / ((-1.0 + N) * N * (1.0 + N))

    return constants.CF * (tmp_const + tmp_S1 * S1 + tmp_S12 * S1**2 + tmp_S2 * S2)


@nb.njit(cache=True)
def gamma_qph(N, nf, cache):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` quark-photon anomalous dimension.

    Implements Eq. (26) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_s^1a_{em}^1)` quark-photon anomalous dimension :math:`\\gamma_{q \\gamma}^{(1,1)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    tmp_const = (
        -2.0
        * (
            4.0
            + 8.0 * N
            + 25.0 * N**2
            + 51.0 * N**3
            + 36.0 * N**4
            + 15.0 * N**5
            + 5.0 * N**6
        )
        / (N**3 * (1.0 + N) ** 3 * (2.0 + N))
    )
    tmp_S1 = 8.0 / N**2
    tmp_S12 = -4.0 * (2.0 + N + N**2) / (N * (1.0 + N) * (2.0 + N))
    tmp_S2 = 4.0 * (2.0 + N + N**2) / (N * (1.0 + N) * (2.0 + N))
    return (
        2.0
        * nf
        * constants.CA
        * constants.CF
        * (tmp_const + tmp_S1 * S1 + tmp_S12 * S1**2 + tmp_S2 * S2)
    )


@nb.njit(cache=True)
def gamma_gph(N):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` gluon-photon anomalous dimension.

    Implements Eq. (27) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    complex
        :math:`O(a_s^1a_{em}^1)` gluon-photon anomalous dimension :math:`\\gamma_{g \\gamma}^{(1,1)}(N)`

    """
    return (
        constants.CF
        * constants.CA
        * (
            8.0
            * (
                -4.0
                + N * (-4.0 + N * (-5.0 + N * (-10.0 + N + 2.0 * N**2 * (2.0 + N))))
            )
        )
        / (N**3 * (1.0 + N) ** 3 * (-2.0 + N + N**2))
    )


@nb.njit(cache=True)
def gamma_phg(N):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` photon-gluon anomalous dimension.

    Implements Eq. (30) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    complex
        :math:`O(a_s^1a_{em}^1)` photon-gluon anomalous dimension :math:`\\gamma_{\\gamma g}^{(1,1)}(N)`

    """
    return constants.TR / constants.CF / constants.CA * constants.NC * gamma_gph(N)


@nb.njit(cache=True)
def gamma_qg(N, nf, cache):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` quark-gluon singlet anomalous dimension.

    Implements Eq. (29) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache


    Returns
    -------
    complex
        :math:`O(a_s^1a_{em}^1)` quark-gluon singlet anomalous dimension
        :math:`\\gamma_{qg}^{(1,1)}(N)`

    """
    return (
        constants.TR
        / constants.CF
        / constants.CA
        * constants.NC
        * gamma_qph(N, nf, cache)
    )


@nb.njit(cache=True)
def gamma_gq(N, cache):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` gluon-quark singlet anomalous dimension.

    Implements Eq. (35) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_s^1a_{em}^1)` gluon-quark singlet anomalous dimension
        :math:`\\gamma_{gq}^{(1,1)}(N)`

    """
    return gamma_phq(N, cache)


@nb.njit(cache=True)
def gamma_phph(nf):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` photon-photon singlet anomalous dimension.

    Implements Eq. (28) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    nf : int
        Number of active flavors

    Returns
    -------
    gamma_gg : complex
        :math:`O(a_s^1a_{em}^1)` photon-photon singlet anomalous dimension
        :math:`\\gamma_{\\gamma \\gamma}^{(1,1)}(N)`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return 4.0 * constants.CF * constants.CA * (nu * constants.eu2 + nd * constants.ed2)


@nb.njit(cache=True)
def gamma_gg():
    r"""Compute the :math:`O(a_s^1a_{em}^1)` gluon-gluon singlet anomalous dimension.

    Implements Eq. (31) of :cite:`deFlorian:2015ujt`.

    Returns
    -------
    complex
        :math:`O(a_s^1a_{em}^1)` gluon-gluon singlet anomalous dimension
        :math:`\\gamma_{gg}^{(1,1)}(N)`

    """
    return 4.0 * constants.TR * constants.NC


@nb.njit(cache=True)
def gamma_nsp(N, cache):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` singlet-like non-singlet anomalous dimension.

    Implements sum of Eqs. (33-34) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    gamma_nsp : complex
        :math:`O(a_s^1a_{em}^1)` singlet-like non-singlet anomalous dimension
        :math:`\\gamma_{ns,+}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    S3 = c.get(c.S3, cache, N)
    S1h = c.get(c.S1h, cache, N)
    S2h = c.get(c.S2h, cache, N)
    S3h = c.get(c.S3h, cache, N)
    S1p1h = c.get(c.S1ph, cache, N)
    S2p1h = c.get(c.S2ph, cache, N)
    S3p1h = c.get(c.S3ph, cache, N)
    g3N = c.get(c.g3, cache, N)
    g3Np2 = c.get(c.g3p2, cache, N)
    result = (
        +32.0 * zeta2 * S1h
        - 32.0 * zeta2 * S1p1h
        + 8.0 / (N + N**2) * S2h
        - 4.0 * S3h
        + (24.0 + 16.0 / (N + N**2)) * S2
        - 32.0 * S3
        - 8.0 / (N + N**2) * S2p1h
        + S1
        * (
            +16.0 * (3.0 / N**2 - 3.0 / (1.0 + N) ** 2 + 2.0 * zeta2)
            - 16.0 * S2h
            - 32.0 * S2
            + 16.0 * S2p1h
        )
        + (
            -8.0
            + N
            * (
                -32.0
                + N
                * (
                    -8.0
                    - 3.0 * N * (3.0 + N) * (3.0 + N**2)
                    - 48.0 * (1.0 + N) ** 2 * zeta2
                )
            )
        )
        / (N**3 * (1.0 + N) ** 3)
        + 32.0 * (g3N + g3Np2)
        + 4.0 * S3p1h
        - 16.0 * zeta3
    )
    return constants.CF * result


@nb.njit(cache=True)
def gamma_nsm(N, cache):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` valence-like non-singlet anomalous dimension.

    Implements difference between Eqs. (33-34) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
    N : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        :math:`O(a_s^1a_{em}^1)` singlet-like non-singlet anomalous dimension
        :math:`\\gamma_{ns,-}^{(1,1)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    S3 = c.get(c.S3, cache, N)
    S1h = c.get(c.S1h, cache, N)
    S2h = c.get(c.S2h, cache, N)
    S3h = c.get(c.S3h, cache, N)
    S1p1h = c.get(c.S1ph, cache, N)
    S2p1h = c.get(c.S2ph, cache, N)
    S3p1h = c.get(c.S3ph, cache, N)
    g3N = c.get(c.g3, cache, N)
    g3Np2 = c.get(c.g3p2, cache, N)
    result = (
        -32.0 * zeta2 * S1h
        - 8.0 / (N + N**2) * S2h
        + (24.0 + 16.0 / (N + N**2)) * S2
        + 8.0 / (N + N**2) * S2p1h
        + S1
        * (
            16.0 * (-1.0 / N**2 + 1.0 / (1.0 + N) ** 2 + 2.0 * zeta2)
            + 16.0 * S2h
            - 32.0 * S2
            - 16.0 * S2p1h
        )
        + (
            72.0
            + N
            * (
                96.0
                - 3.0 * N * (8.0 + 3.0 * N * (3.0 + N) * (3.0 + N**2))
                + 48.0 * N * (1.0 + N) ** 2 * zeta2
            )
        )
        / (3.0 * N**3 * (1.0 + N) ** 3)
        - 32.0 * (g3N + g3Np2)
        + 32.0 * zeta2 * S1p1h
        + 4.0 * S3h
        - 32.0 * S3
        - 4.0 * S3p1h
        - 16.0 * zeta3
    )
    return constants.CF * result


@nb.njit(cache=True)
def gamma_singlet(N, nf, cache):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` singlet sector.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        :math:`O(a_s^1a_{em}^1)` singlet anomalous dimension :math:`\\gamma_{S}^{(1,1)}(N,nf,cache)`

    """
    e2avg, vue2m, vde2m, e2delta = constants.charge_combinations(nf)
    e2_tot = nf * e2avg
    gamma_g_q = gamma_gq(N, cache)
    gamma_ph_q = gamma_phq(N, cache)
    gamma_q_g = gamma_qg(N, nf, cache)
    gamma_q_ph = gamma_qph(N, nf, cache)
    gamma_ns_p = gamma_nsp(N, cache)
    gamma_S_11 = np.array(
        [
            [
                e2_tot * gamma_gg(),
                e2_tot * gamma_gph(N),
                e2avg * gamma_g_q,
                vue2m * gamma_g_q,
            ],
            [
                e2_tot * gamma_phg(N),
                gamma_phph(nf),
                e2avg * gamma_ph_q,
                vue2m * gamma_ph_q,
            ],
            [
                e2avg * gamma_q_g,
                e2avg * gamma_q_ph,
                e2avg * gamma_ns_p,
                vue2m * gamma_ns_p,
            ],
            [
                vde2m * gamma_q_g,
                vde2m * gamma_q_ph,
                vde2m * gamma_ns_p,
                e2delta * gamma_ns_p,
            ],
        ],
        np.complex_,
    )
    return gamma_S_11


@nb.njit(cache=True)
def gamma_valence(N, nf, cache):
    r"""Compute the :math:`O(a_s^1a_{em}^1)` valence sector.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        :math:`O(a_s^1a_{em}^1)` valence anomalous dimension :math:`\\gamma_{V}^{(1,1)}(N,nf,cache)`

    """
    e2avg, vue2m, vde2m, e2delta = constants.charge_combinations(nf)
    gamma_V_11 = np.array(
        [
            [e2avg, vue2m],
            [vde2m, e2delta],
        ],
        np.complex_,
    )
    return gamma_V_11 * gamma_nsm(N, cache)



"""EKO implementation of navigator."""
import os
import webbrowser

import matplotlib.pyplot as plt
import pandas as pd
from banana import cfg as banana_cfg
from banana import navigator as bnav
from banana.data import dfdict
from matplotlib.backends.backend_pdf import PdfPages

from eko import basis_rotation as br

from .. import pdfname
from ..data import db
from ..plots import input_figure, plot_dist

table_objects = bnav.navigator.table_objects
table_objects["o"] = db.Operator


class NavigatorApp(bnav.navigator.NavigatorApp):
    """Navigator base class holding all elementry operations.

    Parameters
    ----------
    cfg : dict
        banana configuration
    mode : str
        mode identifier

    """

    myname = "eko"
    table_objects = table_objects

    def fill_theories(self, theo, obj):
        """Collect important information of the theory record.

        Parameters
        ----------
        theo : dict
            database record
        obj : dict
            to be updated pandas record

        """
        for f in [
            "PTO",
            "ModEv",
            "Q0",
            "Qref",
            "alphas",
            "XIF",
        ]:
            obj[f] = theo[f]
        obj["XIF"] = theo["XIF"]
        obj["mcThr"] = theo["mc"] * theo["kcThr"]
        obj["mbThr"] = theo["mb"] * theo["kbThr"]
        obj["mtThr"] = theo["mt"] * theo["ktThr"]

    def fill_operators(self, op, obj):
        """Collect important information of the operator record.

        Parameters
        ----------
        op : dict
            database record
        obj : dict
            to be updated pandas record

        """
        xgrid = op["interpolation_xgrid"]
        obj["xgrid"] = (
            f"{len(xgrid)}pts: "
            + f"{'log' if op['interpolation_is_log'] else 'x'}"
            + f"^{op['interpolation_polynomial_degree']}"
        )
        obj["mugrid"] = op["mugrid"]
        obj["max_ord"] = op["ev_op_max_order"]
        obj["iters"] = op["ev_op_iterations"]
        obj["skip_ns"] = op["debug_skip_non_singlet"]
        obj["skip_s"] = op["debug_skip_singlet"]
        obj["pol"] = op["polarized"]
        obj["time"] = op["time_like"]

    def fill_cache(self, cac, obj):
        """Collect important information of the cache record.

        Parameters
        ----------
        cac : dict
            database record
        obj : dict
            to be updated pandas record

        """
        vals = cac["result"]["values"]
        q2s = list(vals.keys())
        # assume the vals are homogenous (true for bare eko results) and look
        # only at the first one
        pdfs = len(next(iter(vals.values())))

        obj["operators"] = f"{pdfs} pdfs @ Q^2 {q2s} Gev^2"

        obj["theory"] = cac["t_hash"][: self.hash_len]
        obj["observables"] = cac["o_hash"][: self.hash_len]
        for f in ["pdf", "external"]:
            obj[f] = cac[f]

    def fill_logs(self, lg, obj):
        """Collect important information of the log record.

        Parameters
        ----------
        lg : dict
            database record
        obj : dict
            to be updated pandas record

        """
        q2s = lg["log"].q2s
        crash = lg.get("_crash", None)
        if crash is None:
            obj["q2s"] = f"{q2s}"
        else:
            obj["q2s"] = crash

        obj["theory"] = lg["t_hash"][: self.hash_len]
        obj["ocard"] = lg["o_hash"][: self.hash_len]
        for f in ["pdf", "external"]:
            obj[f] = lg[f]

    def check_log(self, doc_hash, perc_thr=1, abs_thr=1e-6):
        """Check if the log passed the default assertions.

        Parameters
        ----------
        doc_hash : hash
            log hash

        """
        dfds = self.log_as_dfd(doc_hash)
        log = self.get(bnav.l, doc_hash)

        for q2 in dfds:
            for op, df in dfds[q2].items():
                for row in df.iloc:
                    if (
                        abs(row["percent_error"]) > perc_thr
                        and abs(row[f"{log['external']}"] - row["eko"]) > abs_thr
                    ):
                        print(op, row, sep="\n", end="\n\n")

    def plot_pdfs(self, doc_hash):
        """Plot all PDFs at the final scale.

        Parameters
        ----------
        doc_hash : hash
            log hash

        """
        log = self.get(bnav.l, doc_hash)
        dfd = log["log"]
        directory = (
            banana_cfg.cfg["database_path"].parents[0] / f"{log['external']}_bench"
        )

        if not os.path.exists(directory):
            os.makedirs(directory)

        ops_id = log["hash"][: self.hash_len]
        path = directory / f"{ops_id}.pdf"
        print(f"Writing pdf plots to {path}")

        with PdfPages(path) as pp:
            # print setup
            theory = self.get(bnav.t, log["t_hash"][: self.hash_len])
            ops = self.get(bnav.o, log["o_hash"][: self.hash_len])
            firstPage = input_figure(theory, ops, pdf_name=log["pdf"])
            pp.savefig()
            plt.close(firstPage)

            # iterate all pdf
            for q2 in dfd.q2s:
                for op, key in dfd.q2_slice(q2).items():
                    # plot
                    fig = plot_dist(
                        key["x"],
                        key["eko"],
                        key["eko_error"],
                        key[f"{log['external']}"],
                        title=f"x{pdfname(op)}(x,µ_F^2 = {q2} GeV^2)",
                    )
                    pp.savefig()
                    plt.close(fig)

        return path

    def display_pdfs(self, doc_hash):
        """Display PDF generated by ekomark.navigator.navigator.plot_pdfs.

        Parameters
        ----------
        doc_hash : hash
            log hash

        """
        log = self.get(bnav.l, doc_hash)
        directory = (
            banana_cfg.cfg["database_path"].parents[0] / f"{log['external']}_bench"
        )

        if not directory.exists():
            directory.mkdir(parents=True)

        path = None
        for plot_file in directory.iterdir():
            if plot_file.stem[: len(doc_hash)] == doc_hash:
                path = plot_file
                break

        if path is None:
            print("Drawing the plots...")
            path = self.plot_pdfs(doc_hash)

        webbrowser.open(str(path.absolute()))

    @staticmethod
    def is_valid_physical_object(name):
        """Check if is actual name.

        Parameters
        ----------
        name : str
            name

        Returns
        -------
        bool :
            is actual name?

        """
        return name in br.evol_basis or name in br.flavor_basis_names

    def compare_external(self, dfd1, dfd2):
        """Compare two results in the cache.

        It's taking two results from external benchmarks and compare them in a
        single table.

        Parameters
        ----------
        dfd1 : dict or hash
            if hash the doc_hash of the cache to be loaded
        dfd2 : dict or hash
            if hash the doc_hash of the cache to be loaded

        """
        # load json documents
        id1, cache1 = self.load_dfd(dfd1, self.cache_as_dfd)
        id2, cache2 = self.load_dfd(dfd2, self.cache_as_dfd)

        if cache1.external == cache2.external:
            cache1.external = f"{cache1.external}1"
            cache2.external = f"{cache2.external}2"

        # print head
        cache_diff = dfdict.DFdict()
        msg = f"**Comparing** id: `{id1}` - id: `{id2}`, in table *cache*"
        cache_diff.print(msg, "-" * len(msg), sep="\n")
        cache_diff.print(f"- *{cache1.external}*: `{id1}`")
        cache_diff.print(f"- *{cache2.external}*: `{id2}`")
        cache_diff.print()

        # check x
        for x in cache1["target_xgrid"]:
            if x not in cache2["target_xgrid"]:
                raise ValueError(f"{x}: not matching in x")

        table_out = dfdict.DFdict()
        for q2, pdfs1 in cache1["values"].items():
            # check q2
            if q2 not in cache2["values"].keys():
                raise ValueError(f"{q2}: not matching in q2")

            pdfs2 = cache2["values"][q2]
            # check pdfs
            for pdf in pdfs1.T.keys():
                if pdf not in pdfs2.keys():
                    raise ValueError(f"{pdf}: pdf not matching")
                # skip the photon always (just for the time being)
                if pdf == "ph":
                    continue

                tab = {
                    "x": cache1["target_xgrid"][0],
                    cache1.external: pdfs1[pdf],
                    cache2.external: pdfs2[pdf],
                    "percent_error": (pdfs1[pdf] / pdfs2[pdf] - 1.0) * 100,
                }
                table_out[pdf] = pd.DataFrame.from_dict(tab)

            # dump results' table
            cache_diff[q2] = table_out

        return cache_diff
